import requests

from logging import Logger
from requests import Response
from typing import Dict

from helper import retry, json_serialize

from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import ErrorDownloadingLoaders
from application.util import ErrorResponseUtil, settings


class LoaderUpdaterAPIRepository(object):

    def __init__(self, logger, configs):
        # type: (Logger, Configurations) -> None

        self.logger = logger
        self.configs = configs
        self.update_type = UpdateType.loader
        self.local_configs = configs.updaters[self.update_type.name]
        self.response_util = ErrorResponseUtil()
        self.error_response = None

    def download_loaders(self, file_path, params):
        # type: (str, Dict) -> None

        url = self._get_download_url(params)
        headers = self._get_zip_headers()
        response = retry(
                logger=self.logger,
                method_name="download_loaders",
                n=3,
                sleep_seconds=5,
                trying_to_do=lambda: requests.request("GET", url=url, headers=headers, timeout=60),
                is_success=lambda resp: resp.status_code == 302,
                success=lambda resp: resp,
                failed=lambda resp: self._get_error_response(resp)
        )

        if response is None:
            raise ErrorDownloadingLoaders(self.error_response)

        package = self._download_package(response.content.strip('\"'))
        self._check_has_file(package)
        self._save_package(file_path, package)

    @staticmethod
    def _save_package(file_path, package):
        # type: (str, Response) -> None

        with open(file_path, "wb") as zip_object:
            for chunk in package.iter_content(chunk_size=128):
                zip_object.write(chunk)

    @staticmethod
    def _download_package(download_url):
        # type: (str) -> Response

        return requests.get(download_url, allow_redirects=True)

    @staticmethod
    def _get_url_params(params):
        # type: (Dict) -> str
        url = ""
        for key, value in params.items():
            if not url:
                url += "?"
            else:
                url += "&"
            url += "{}={}".format(key, value)
        return url

    def _get_download_url(self, params):
        # type: (Dict) -> str

        end_point_url = self.local_configs.endpoints.download_loaders.format(storeId=self.configs.store_id)
        params_url = self._get_url_params(params)
        return self.configs.base_url + end_point_url + params_url

    def _get_zip_headers(self):
        # type: () -> Dict[str, str]

        return {"X-Api-Key": self.configs.api_key,
                "Content-type": "application/zip"}

    def _get_error_response(self, resp):
        # type: (Response) -> None

        self.logger.error("Error downloading loaders: {}".format(json_serialize(resp)))
        self.error_response = self.response_util.get_error_response(resp, "$STORE_NOT_FOUND")

    @staticmethod
    def _check_has_file(package):
        # type: (Response) -> None

        if package is not None and package.status_code == 403:
            settings.set_update_control("loaders", "$LOADERS_NOT_FOUND")
            raise ErrorDownloadingLoaders("$LOADERS_NOT_FOUND")



# rnc.py - functions for handling Dominican Republic tax registration
# coding: utf-8
#
# Copyright (C) 2015-2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

# Development of this functionality was funded by iterativo | https://iterativo.do

"""RNC (Registro Nacional del Contribuyente, Dominican Republic tax number).

The RNC is the Dominican Republic taxpayer registration number for
institutions. The number consists of 9 digits.

>>> validate('1-01-85004-3')
'101850043'
>>> validate('1018A0043')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('101850042')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('131246796')
'1-31-24679-6'
"""

import json

from stdnum.exceptions import *
from stdnum.util import clean, get_soap_client, isdigits


# list of RNCs that do not match the checksum but are nonetheless valid
whitelist = set('''
101581601 101582245 101595422 101595785 10233317 131188691 401007374
501341601 501378067 501620371 501651319 501651823 501651845 501651926
501656006 501658167 501670785 501676936 501680158 504654542 504680029
504681442 505038691
'''.split())


dgii_wsdl = 'https://www.dgii.gov.do/wsMovilDGII/WSMovilDGII.asmx?WSDL'
"""The WSDL URL of DGII validation service."""


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def calc_check_digit(number):
    """Calculate the check digit."""
    weights = (7, 9, 8, 6, 5, 4, 3, 2)
    check = sum(w * int(n) for w, n in zip(weights, number)) % 11
    return str((10 - check) % 9 + 1)


def validate(number):
    """Check if the number provided is a valid RNC."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if number in whitelist:
        return number
    if len(number) != 9:
        raise InvalidLength()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid RNC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join((number[:1], number[1:3], number[3:-1], number[-1]))


def _convert_result(result):  # pragma: no cover
    """Translate SOAP result entries into dicts."""
    translation = {
        'RGE_RUC': 'rnc',
        'RGE_NOMBRE': 'name',
        'NOMBRE_COMERCIAL': 'commercial_name',
        'CATEGORIA': 'category',
        'REGIMEN_PAGOS': 'payment_regime',
        'ESTATUS': 'status',
        'RNUM': 'result_number',
    }
    return dict(
        (translation.get(key, key), value)
        for key, value in json.loads(result.replace('\n', '\\n').replace('\t', '\\t')).items())


def check_dgii(number, timeout=30):  # pragma: no cover
    """Lookup the number using the DGII online web service.

    This uses the validation service run by the the Dirección General de
    Impuestos Internos, the Dominican Republic tax department to lookup
    registration information for the number. The timeout is in seconds.

    Returns a dict with the following structure::

        {
            'rnc': '123456789',     # The requested number
            'name': 'The registered name',
            'commercial_name': 'An additional commercial name',
            'status': '2',          # 1: inactive, 2: active
            'category': '0',        # always 0?
            'payment_regime': '2',  # 1: N/D, 2: NORMAL, 3: PST
        }

    Will return None if the number is invalid or unknown."""
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the online service
    number = compact(number)
    client = get_soap_client(dgii_wsdl, timeout)
    result = client.GetContribuyentes(
        value=number,
        patronBusqueda=0,   # search type: 0=by number, 1=by name
        inicioFilas=1,      # start result (1-based)
        filaFilas=1,        # end result
        IMEI='')
    if result and 'GetContribuyentesResult' in result:
        result = result['GetContribuyentesResult']  # PySimpleSOAP only
    if result == '0':
        return
    result = [x for x in result.split('@@@')]
    return _convert_result(result[0])


def search_dgii(keyword, end_at=10, start_at=1, timeout=30):  # pragma: no cover
    """Search the DGII online web service using the keyword.

    This uses the validation service run by the the Dirección General de
    Impuestos Internos, the Dominican Republic tax department to search the
    registration information using the keyword.

    The number of entries returned can be tuned with the `end_at` and
    `start_at` arguments. The timeout is in seconds.

    Returns a list of dicts with the following structure::

        [
            {
                'rnc': '123456789',     # The found number
                'name': 'The registered name',
                'commercial_name': 'An additional commercial name',
                'status': '2',          # 1: inactive, 2: active
                'category': '0',        # always 0?
                'payment_regime': '2',  # 1: N/D, 2: NORMAL, 3: PST
                'result_number': '1',   # index of the result
            },
            ...
        ]

    Will return an empty list if the number is invalid or unknown."""
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the online service
    client = get_soap_client(dgii_wsdl, timeout)
    results = client.GetContribuyentes(
        value=keyword,
        patronBusqueda=1,       # search type: 0=by number, 1=by name
        inicioFilas=start_at,   # start result (1-based)
        filaFilas=end_at,       # end result
        IMEI='')
    if results and 'GetContribuyentesResult' in results:
        results = results['GetContribuyentesResult']  # PySimpleSOAP only
    if results == '0':
        return []
    return [_convert_result(result) for result in results.split('@@@')]



from .. import Waiter
from messagebus import Event
from typing import List


class ResolvedWaiter(Waiter):
    def __init__(self, sync_id, events):
        # type: (str, List[Event]) -> None
        self.sync_id = sync_id
        self.events = events

    def get_sync_id(self):
        return self.sync_id

    def get_events(self):
        return self.events



class FiscalIncrementException(Exception):
    pass


class FiscalGetIdException(Exception):
    pass


class FiscalParametersException(Exception):
    pass


class FiscalValidation(Exception):
    pass



# bitcoin.py - functions for handling Bitcoin addresses
#
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Bitcoin address.

A Bitcoin address is an identifier that is used as destination in a Bitcoin
transaction. It is based on a hash of the public portion of a key pair.

There are currently three address formats in use:

* P2PKH: pay to pubkey hash
* P2SH: pay to script hash
* Bech32

More information:

* https://en.bitcoin.it/wiki/Address

>>> validate('1NEDqZPvTWRaoho48qXuLLsrYomMXPABfD')
'1NEDqZPvTWRaoho48qXuLLsrYomMXPABfD'
>>> validate('BC1QARDV855YJNGSPVXUTTQ897AQCA3LXJU2Y69JCE')
'bc1qardv855yjngspvxuttq897aqca3lxju2y69jce'
>>> validate('1NEDqZPvTWRaoho48qXuLLsrYomMXPABfX')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

import hashlib
import struct
from functools import reduce

from stdnum.exceptions import *
from stdnum.util import clean


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' ').strip()
    if number[:3].lower() == 'bc1':
        number = number.lower()
    return number


# Base58 encoding character set as used in Bitcoin addresses
_base58_alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def b58decode(s):
    """Decode a Base58 encoded string to a bytestring."""
    value = reduce(lambda a, c: a * 58 + _base58_alphabet.index(c), s, 0)
    result = b''
    while value >= 256:
        value, mod = divmod(value, 256)
        result = struct.pack('B', mod) + result
    result = struct.pack('B', value) + result
    return struct.pack('B', 0) * (len(s) - len(s.lstrip('1'))) + result


# Bech32 character set as used in Bitcoin addresses
_bech32_alphabet = 'qpzry9x8gf2tvdw0s3jn54khce6mua7l'


# The Bech32 generator tests and values for checksum calculation
_bech32_generator = (
    (1 << 0, 0x3b6a57b2), (1 << 1, 0x26508e6d), (1 << 2, 0x1ea119fa),
    (1 << 3, 0x3d4233dd), (1 << 4, 0x2a1462b3))


def bech32_checksum(values):
    """Calculate the Bech32 checksum."""
    chk = 1
    for value in values:
        top = chk >> 25
        chk = (chk & 0x1ffffff) << 5 | value
        for t, v in _bech32_generator:
            chk ^= v if top & t else 0
    return chk


def b32decode(data):
    """Decode a list of Base32 values to a bytestring."""
    acc, bits = 0, 0
    result = b''
    for value in data:
        acc = ((acc << 5) | value) & 0xfff
        bits += 5
        if bits >= 8:
            bits -= 8
            result = result + struct.pack('B', (acc >> bits) & 0xff)
    if bits >= 5 or acc & ((1 << bits) - 1):
        raise InvalidComponent()
    return result


def _expand_hrp(hrp):
    """Convert the human-readable part to format for checksum calculation."""
    return [ord(c) >> 5 for c in hrp] + [0] + [ord(c) & 31 for c in hrp]


def validate(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    number = compact(number)
    if number.startswith('1') or number.startswith('3'):
        # P2PKH (pay to pubkey hash) or P2SH (pay to script hash) address
        if not all(x in _base58_alphabet for x in number):
            raise InvalidFormat()
        address = b58decode(number)
        if len(address) != 25:
            raise InvalidLength()
        if hashlib.sha256(hashlib.sha256(address[:-4]).digest()).digest()[:4] != address[-4:]:
            raise InvalidChecksum()
    elif number.startswith('bc1'):
        # Bech32 type address
        if not all(x in _bech32_alphabet for x in number[3:]):
            raise InvalidFormat()
        if len(number) < 11 or len(number) > 90:
            raise InvalidLength()
        data = [_bech32_alphabet.index(x) for x in number[3:]]
        if bech32_checksum(_expand_hrp('bc') + data) != 1:
            raise InvalidChecksum()
        witness_version = data[0]
        witness_program = b32decode(data[1:-6])
        if witness_version > 16:
            raise InvalidComponent()
        if len(witness_program) < 2 or len(witness_program) > 40:
            raise InvalidLength()
        if witness_version == 0 and len(witness_program) not in (20, 32):
            raise InvalidLength()
    else:
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import re

from logging import Logger
from typing import (
    List,
    Optional,
    Union,
)

from production.box import ProductionBox
from production.model import Item
from production.repository import ProductRepository

from ._JitItemFilterBox import JitItemFilterBox


class JitSequencerItemFilterBox(JitItemFilterBox):
    def __init__(
            self,
            name,                   # type: str
            sons,                   # type: Optional[Union[List[ProductionBox]], ProductionBox]
            repository,             # type: ProductRepository
            allowed_jit_lines,      # type: Union[str, List[str]]
            forbidden_jit_lines,    # type: Union[str, List[str]]
            logger=None,            # type: Logger
    ):
        # type: (...) -> None

        super(JitSequencerItemFilterBox, self).__init__(
                name,
                sons,
                repository,
                allowed_jit_lines,
                forbidden_jit_lines,
                logger,
        )

    def change_order(self, order):
        filtered_items = []
        for item in order.items:
            self.handle_item(item, filtered_items)

        new_items = []
        for item in filtered_items:
            self._filter_sequencer_item(item, new_items)

        order.items = new_items
        return order

    def _filter_sequencer_item(self, item, new_items):
        # type: (Item, List[Item]) ->  None

        if self._must_filter_part_code(item):
            return

        if len(item.items) > 0:
            sons = []
            for son in item.items:
                if not son.displayed_jit and item.displayed_jit:
                    son.displayed_jit = item.displayed_jit

                self._filter_sequencer_item(son, sons)

            if item.item_type in ("OPTION",) and not sons:
                return

            item.items = sons

        new_items.append(item)

    def _must_filter_part_code(self, item):
        # type: (Item) -> bool

        part_code_jit = item.jit_lines
        if self.allowed_jit_lines is None or not part_code_jit or item.has_tag("dont-need-cook"):
            return False

        for jit in self.allowed_jit_lines:
            if jit not in part_code_jit:
                continue

            item.displayed_jit.append(jit)
            jit_sequence = [x for x in re.split('[><$]', part_code_jit) if jit in x][0]
            jit_sequence_list = jit_sequence.split("-")
            jit_index = 0
            for jit_elem in jit_sequence_list:
                if jit in jit_elem.split("|"):
                    break

                jit_index += 1

            if jit_index == 0:
                return False

            previous_jit = jit_sequence_list[jit_index-1]
            necessary_jit_lines = re.findall('(\\w+)', previous_jit)
            done_tag = self._check_all_tag_in_item(item, necessary_jit_lines, ["done", "served"])
            return done_tag is None

        return True

    @staticmethod
    def _check_all_tag_in_item(item, necessary_jit_lines, tag_types):
        last_tag_name = None
        for jit in necessary_jit_lines:
            for tag_type in tag_types:
                tag_name = "{}-{}".format(tag_type, jit.lower())
                if tag_name in item.tags:
                    last_tag_name = tag_name
                    break
            else:
                return None

        return last_tag_name



from report.command import AlignCommand

from .._CommandProcessor import CommandProcessor


class AlignCommandProcessor(CommandProcessor):
    def get_bytes(self, command):
        if not isinstance(command, AlignCommand):
            return ""
        
        ret = b"\x1b\x61"
        if command.alignment == AlignCommand.Alignment.left:
            ret += "\x00"
        elif command.alignment == AlignCommand.Alignment.center:
            ret += "\x01"
        elif command.alignment == AlignCommand.Alignment.right:
            ret += "\x02"
        else:
            return ""

        return ret



# nid.py - functions for handling Mauritian national ID numbers
# coding: utf-8
#
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ID number (Mauritian national identifier).

The Mauritian national ID number is a unique 14 alphanumeric identifier
assigned at birth to identify individuals. It is displayed on the National
Identity Card.

The number consists of one alphabetic character and thirteen digits:

* the first character of the person's surname at birth
* 2 digits for day of birth
* 2 digits for month of birth
* 2 digits for year of birth
* 6 digit unique id
* a check digit

More information:

* https://mnis.govmu.org/English/ID%20Card/Pages/default.aspx
"""

import datetime
import re

from stdnum.exceptions import *
from stdnum.util import clean


_nid_re = re.compile('^[A-Z][0-9]+[0-9A-Z]$')


# characters used for checksum calculation
_alphabet = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'


def compact(number):
    """Convert the number to the minimal representation. This strips
    surrounding whitespace and separation dash."""
    return clean(number, ' ').upper().strip()


def calc_check_digit(number):
    """Calculate the check digit for the number."""
    check = sum((14 - i) * _alphabet.index(n)
                for i, n in enumerate(number[:13]))
    return _alphabet[(17 - check) % 17]


def _get_date(number):
    """Convert the part of the number that represents a date into a
    datetime. Note that the century may be incorrect."""
    day = int(number[1:3])
    month = int(number[3:5])
    year = int(number[5:7])
    try:
        return datetime.date(year + 2000, month, day)
    except ValueError:
        raise InvalidComponent()


def validate(number):
    """Check if the number is a valid ID number."""
    number = compact(number)
    if len(number) != 14:
        raise InvalidLength()
    if not _nid_re.match(number):
        raise InvalidFormat()
    if calc_check_digit(number) != number[-1]:
        raise InvalidChecksum()
    _get_date(number)
    return number


def is_valid(number):
    """Check if the number provided is a valid RFC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from production.model import Course
from production.repository import ProductRepository
from typing import List


class GetCoursesInteractor(object):
    def __init__(self, product_repository):
        # type: (ProductRepository) -> None

        self.product_repository = product_repository

    def execute(self):
        # type: () -> List[Course]

        return self.product_repository.get_all_courses()



# __init__.py - main module
# coding: utf-8
#
# Copyright (C) 2010-2023 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Parse, validate and reformat standard numbers and codes.

This library offers functions for parsing, validating and reformatting
standard numbers and codes in various formats.

All modules implement a common interface:

>>> from stdnum import isbn
>>> isbn.validate('978-9024538270')
'9789024538270'
>>> isbn.validate('978-9024538271')
Traceback (most recent call last):
    ...
InvalidChecksum: ...

Apart from the validate() function, many modules provide extra
parsing, validation, formatting or conversion functions.
"""

from stdnum.util import get_cc_module


__all__ = ('get_cc_module', '__version__')

# the version number of the library
__version__ = '1.19'



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2007, 2009-2011 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import struct
import base64

import dns.exception
import dns.dnssec
import dns.rdata
import dns.tokenizer

_ctype_by_value = {
    1: 'PKIX',
    2: 'SPKI',
    3: 'PGP',
    253: 'URI',
    254: 'OID',
}

_ctype_by_name = {
    'PKIX': 1,
    'SPKI': 2,
    'PGP': 3,
    'URI': 253,
    'OID': 254,
}


def _ctype_from_text(what):
    v = _ctype_by_name.get(what)
    if v is not None:
        return v
    return int(what)


def _ctype_to_text(what):
    v = _ctype_by_value.get(what)
    if v is not None:
        return v
    return str(what)


class CERT(dns.rdata.Rdata):

    """CERT record

    @ivar certificate_type: certificate type
    @type certificate_type: int
    @ivar key_tag: key tag
    @type key_tag: int
    @ivar algorithm: algorithm
    @type algorithm: int
    @ivar certificate: the certificate or CRL
    @type certificate: string
    @see: RFC 2538"""

    __slots__ = ['certificate_type', 'key_tag', 'algorithm', 'certificate']

    def __init__(self, rdclass, rdtype, certificate_type, key_tag, algorithm,
                 certificate):
        super(CERT, self).__init__(rdclass, rdtype)
        self.certificate_type = certificate_type
        self.key_tag = key_tag
        self.algorithm = algorithm
        self.certificate = certificate

    def to_text(self, origin=None, relativize=True, **kw):
        certificate_type = _ctype_to_text(self.certificate_type)
        return "%s %d %s %s" % (certificate_type, self.key_tag,
                                dns.dnssec.algorithm_to_text(self.algorithm),
                                dns.rdata._base64ify(self.certificate))

    @classmethod
    def from_text(cls, rdclass, rdtype, tok, origin=None, relativize=True):
        certificate_type = _ctype_from_text(tok.get_string())
        key_tag = tok.get_uint16()
        algorithm = dns.dnssec.algorithm_from_text(tok.get_string())
        if algorithm < 0 or algorithm > 255:
            raise dns.exception.SyntaxError("bad algorithm type")
        chunks = []
        while 1:
            t = tok.get().unescape()
            if t.is_eol_or_eof():
                break
            if not t.is_identifier():
                raise dns.exception.SyntaxError
            chunks.append(t.value.encode())
        b64 = b''.join(chunks)
        certificate = base64.b64decode(b64)
        return cls(rdclass, rdtype, certificate_type, key_tag,
                   algorithm, certificate)

    def to_wire(self, file, compress=None, origin=None):
        prefix = struct.pack("!HHB", self.certificate_type, self.key_tag,
                             self.algorithm)
        file.write(prefix)
        file.write(self.certificate)

    @classmethod
    def from_wire(cls, rdclass, rdtype, wire, current, rdlen, origin=None):
        prefix = wire[current: current + 5].unwrap()
        current += 5
        rdlen -= 5
        if rdlen < 0:
            raise dns.exception.FormError
        (certificate_type, key_tag, algorithm) = struct.unpack("!HHB", prefix)
        certificate = wire[current: current + rdlen].unwrap()
        return cls(rdclass, rdtype, certificate_type, key_tag, algorithm,
                   certificate)



from _AposA8FontSize import AposA8FontSize
from _SatAposA8FiscalPrinter import SatAposA8FiscalPrinter
from _NfceAposA8FiscalPrinter import NfceAposA8FiscalPrinter
from _AposA8PrintLine import AposA8PrintLine
from _AposA8CommandType import AposA8CommandType
from _AposA8Align import AposA8Align
from _AposA8Constants import AposA8Constants


__all__ = [
    "AposA8FontSize",
    "SatAposA8FiscalPrinter",
    "NfceAposA8FiscalPrinter",
    "AposA8PrintLine",
    "AposA8CommandType",
    "AposA8Align",
    "AposA8Constants",
]



import logging
from datetime import datetime, timedelta
from threading import Lock

from sysactions import action
from application.interactor import NDiscountInteractor, NCouponInteractor

coupons_available_cache = None
last_cache_time = None
lock = Lock()
logger = logging.getLogger("PosActions")


@action
def is_coupon_component_available(pos_id, seconds_to_renew_cache=3600):
    is_coupon_component_connected(pos_id, seconds_to_renew_cache=seconds_to_renew_cache)

    global coupons_available_cache
    return str(coupons_available_cache).lower()


def is_coupon_component_connected(pos_id, seconds_to_renew_cache=3600):
    global last_cache_time
    global coupons_available_cache

    with lock:
        now = datetime.now()
        seconds_to_renew = _get_seconds_to_renew(seconds_to_renew_cache)
        if coupons_available_cache is None or _reached_seconds_to_renew_cache(now, seconds_to_renew):
            coupons_available_cache = _test_connection_coupon_discount_components(pos_id)
            last_cache_time = now
        return coupons_available_cache


def _get_seconds_to_renew(seconds_to_renew_cache):
    seconds_to_renew = 180
    try:
        seconds_to_renew = int(seconds_to_renew_cache)
    except TypeError as _:
        pass
    return seconds_to_renew


def _reached_seconds_to_renew_cache(now, seconds_to_renew_cache):
    return not last_cache_time or now > last_cache_time + timedelta(seconds=seconds_to_renew_cache)


def _test_connection_coupon_discount_components(pos_id):
    coupon_interactor = NCouponInteractor()
    discount_interactor = NDiscountInteractor()
    test_coupon = coupon_interactor.test_connection(pos_id)
    test_discount = discount_interactor.test_connection()
    return test_coupon and test_discount



import logging

from application.repository.api import APIRepository
from typing import Optional

logger = logging.getLogger(__name__)


class GetAndLockCouponInteractor(object):
    def __init__(
        self,            # type: GetAndLockCouponInteractor
        api_repository,  # type: APIRepository
    ):
        # type: (...) -> None

        self.api_repository = api_repository

    def get_and_lock_voucher(
        self,       # type: GetAndLockCouponInteractor
        pos_id,     # type: int
        coupon_id,  # type: str
    ):
        # type: (...) -> Optional[str]

        logger.info("Requesting and locking benefit for couponId: {}; PosId: {}".format(coupon_id, pos_id))
        benefit = self.api_repository.retrieve_benefit(
            pos_id=pos_id,
            coupon_id=coupon_id,
        )
        logger.info("Retrieved benefit:\n{}".format(benefit or ""))
        return benefit



from datetime import datetime

from production.model import ProdStates, ProductionOrder
from ._OrderChangerProductionBox import OrderChangerProductionBox


class ReprocessOrdersBox(OrderChangerProductionBox):
    def __init__(self, name, sons, publish_scheduler, reprocess_time=None, reprocess_display_time=None, logger=None):
        super(ReprocessOrdersBox, self).__init__(name, sons, logger)
        self.publish_scheduler = publish_scheduler
        self.reprocess_time = reprocess_time
        self.reprocess_display_time = reprocess_display_time if self.reprocess_time is None else None

    def change_order(self, order):
        # type: (ProductionOrder) -> ProductionOrder

        if order.prod_state != ProdStates.INVALID:
            if self.reprocess_time is not None:
                self.publish_scheduler.schedule_publish(order.order_id, self.reprocess_time)

            if self.reprocess_display_time is not None and order.display_time not in [None, "", " "]:
                display_time = datetime.strptime(order.display_time[0:19], "%Y-%m-%dT%H:%M:%S")
                current_time = datetime.now()

                seconds_to_reprocess = self.reprocess_display_time - (current_time - display_time).total_seconds()
                if seconds_to_reprocess < 0:
                    seconds_to_reprocess = 1
                self.publish_scheduler.schedule_publish(order.order_id, seconds_to_reprocess)

        return order



# upn.py - functions for handling English UPNs
#
# Copyright (C) 2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""UPN (English Unique Pupil Number).

The Unique Pupil Number (UPN) is a 13-character code that identifies pupils
in English state schools and is designed to aid tracking pupil progress
through the school system.

The number consists of a check letter, a 3-digit LA (Local Authority) number
for the issuing school, a 4-digit DfE number (School Establishment Number),
2 digits for the issue year and 3 digits for a serial number. Temporary
numbers have a 2-digit serial and a letter.

More information:

* https://www.gov.uk/government/publications/unique-pupil-numbers

>>> validate('B801200005001')
'B801200005001'
>>> validate('A801200005001')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('X80120000A001')  # middle part must be numeric
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('E000200005001')  # LA number must be known
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# The allowed characters in an UPN.
_alphabet = 'ABCDEFGHJKLMNPQRTUVWXYZ0123456789'


# The known values for the LA (Local Authority) number.
# https://www.gov.uk/government/statistics/new-local-authority-codes-january-2011
_la_numbers = set((
    201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 301,
    302, 303, 304, 305, 306, 307, 308, 309, 310, 311, 312, 313, 314, 315,
    316, 317, 318, 319, 320, 330, 331, 332, 333, 334, 335, 336, 340, 341,
    342, 343, 344, 350, 351, 352, 353, 354, 355, 356, 357, 358, 359, 370,
    371, 372, 373, 380, 381, 382, 383, 384, 390, 391, 392, 393, 394, 420,
    800, 801, 802, 803, 805, 806, 807, 808, 810, 811, 812, 813, 815, 816,
    821, 822, 823, 825, 826, 830, 831, 835, 836, 837, 840, 841, 845, 846,
    850, 851, 852, 855, 856, 857, 860, 861, 865, 866, 867, 868, 869, 870,
    871, 872, 873, 874, 876, 877, 878, 879, 880, 881, 882, 883, 884, 885,
    886, 887, 888, 889, 890, 891, 892, 893, 894, 895, 896, 908, 909, 916,
    919, 921, 925, 926, 928, 929, 931, 933, 935, 936, 937, 938))


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').upper().strip()


def calc_check_digit(number):
    """Calculate the check digit for the number."""
    check = sum(i * _alphabet.index(n)
                for i, n in enumerate(number[-12:], 2)) % 23
    return _alphabet[check]


def validate(number):
    """Check if the number is a valid UPN. This checks length, formatting and
    check digits."""
    number = compact(number)
    if len(number) != 13:
        raise InvalidLength()
    if not isdigits(number[1:-1]) or number[-1] not in _alphabet:
        raise InvalidFormat()
    if int(number[1:4]) not in _la_numbers:
        raise InvalidComponent()
    if calc_check_digit(number[1:]) != number[0]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid UPN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# -*- coding: utf-8 -*-

import json

from sysactions import action, get_model, get_posot, show_messagebox, close_asynch_dialog

from utilfunctions import compare_objects

from ..customer import Customer
from ..customer.customer_model import AddCustomerRequest

dialog_id = None
delivery_keys = [
    "ADDRESS_REFERENCE",
    "CITY",
    "COMPLEMENT",
    "NEIGHBORHOOD",
    "POSTAL_CODE",
    "STATE",
    "STREET_NAME",
    "STREET_NUMBER",
]


def insert_or_update_customer(
    pos_id,                # type: str
    customer_json,         # type: dict
    update_order=False,    # type: bool
):
    # type: (...) -> None

    model = get_model(pos_id)
    pos_ot = get_posot(model)

    customer_request = _create_customer_request(customer_json)
    if update_order:
        _update_order_custom_properties(
            address_data_load=customer_json,
            model=model,
            pos_id=pos_id,
            pos_ot=pos_ot,
        )

    if not customer_request.phone:
        return

    customer = Customer().get_customer(phone=customer_request.phone)
    if not customer:
        message = "$CUSTOMER_WILL_BE_INSERTED"
        confirmation = show_messagebox(
            posid=pos_id,
            message=message,
            timeout=720000,
            buttons="$OK|$CANCEL",
        )
        if confirmation == 1:
            return
        Customer().add_customer(add_customer_request=customer_request)
    else:
        same_customer = compare_objects(
            obj1=customer_request,
            obj2=_create_customer_request_to_compare(customer=json.loads(customer)),
        )
        if not same_customer:
            message = "$CUSTOMER_WILL_BE_UPDATED"
            confirmation = show_messagebox(
                posid=pos_id,
                message=message,
                timeout=720000,
                buttons="$YES|$NO",
            )
            if confirmation == 1:
                return
        Customer().update_customer(add_customer_request=customer_request)


def insert_customer_and_save_on_sale(
    pos_id,                    # type: str
    customer_json,             # type: dict
    create_new_order=False,    # type: bool
):
    # type: (...) -> None

    selected_option = customer_json.get('SELECTED_OPTION', None)
    if selected_option is not None:
        customer_json.pop('SELECTED_OPTION')

    insert_or_update_customer(
        pos_id=pos_id,
        customer_json=customer_json,
        update_order=not create_new_order,
    )

    if create_new_order and selected_option == 2:
        from posactions import doSale

        doSale(
            pos_id=pos_id,
            item_id="",
            sale_type="DELIVERY",
            delivery_info_already_asked=True,
        )
        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)
        _update_order_custom_properties(
            address_data_load=customer_json,
            model=model,
            pos_id=pos_id,
            pos_ot=pos_ot,
        )


def _create_customer_request(address_data_load):
    add_customer_request = AddCustomerRequest(
            address_data_load["CUSTOMER_PHONE"],
            address_data_load["CUSTOMER_DOC"],
            address_data_load["CUSTOMER_NAME"],
            address_data_load["POSTAL_CODE"],
            address_data_load["STATE"],
            address_data_load["CITY"],
            address_data_load["STREET_NAME"],
            address_data_load["NEIGHBORHOOD"],
            address_data_load["STREET_NUMBER"],
            address_data_load["COMPLEMENT"],
            address_data_load["ADDRESS_REFERENCE"]
    )

    return add_customer_request


def _create_customer_request_to_compare(customer):
    add_customer_request = AddCustomerRequest(
            customer["phone"],
            customer["document"],
            customer["name"],
            customer["address"]["zip_code"],
            customer["address"]["state"],
            customer["address"]["city"],
            customer["address"]["street"],
            customer["address"]["neighborhood"],
            customer["address"]["number"],
            customer["address"]["complement"],
            customer["address"]["reference_point"]
    )
    return add_customer_request


@action
def zip_code_search(pos_id):
    global dialog_id
    dialog_id = show_messagebox(pos_id, "$SEARCHING_ZIP_CODE", buttons="", asynch=True, timeout=50000)


@action
def zip_code_search_finished(pos_id, error="false"):
    close_asynch_dialog(pos_id, dialog_id)
    if error == "true":
        show_messagebox(pos_id, "$ERROR_SEARCHING_ZIP_CODE", buttons="OK")


def _update_order_custom_properties(address_data_load, model, pos_id, pos_ot):
    for key, value in address_data_load.iteritems():
        if value in ("", "None", None):
            if model.find('.//OrderProperty[@key="{}"]'.format(key)) is not None:
                pos_ot.setOrderCustomProperty(key, 'clear')
        else:
            pos_ot.setOrderCustomProperty(key, value.encode('utf-8'))

    _update_order_properties(pos_id, pos_ot, model)


def _update_order_properties(pos_id, pos_ot, model):
    sale_type = model.find(".//Order").get("saleType")
    pos_ot.updateOrderProperties(pos_id, saletype=sale_type)


def has_delivery_info(
    pos_id,    # type: str
):
    # type: (...) -> bool

    model = get_model(posid=pos_id)
    order = model.find("CurrentOrder/Order")
    if order is None:
        return False

    for prop in order.findall('.//OrderProperty'):
        if prop.get('key') in delivery_keys:
            return True
    return False




from _UpdateGoalsProcessor import UpdateGoalsProcessor
from _UpdateTotalSoldProcessor import UpdateTotalSoldProcessor
from _UpdateTotalSoldOrderProcessor import UpdateTotalSoldOrderProcessor
from _UpdateTotalSoldTableProcessor import UpdateTotalSoldTableProcessor
from _GetSalesChallengeProcessor import GetSalesChallengeProcessor
from _InsertSalesChallengeProcessor import InsertSalesChallengeProcessor
from _DeleteSalesChallengeProcessor import DeleteSalesChallengeProcessor
from _UpdateSalesChallengeProcessor import UpdateSalesChallengeProcessor



# vat.py - functions for handling Cypriot VAT numbers
# coding: utf-8
#
# Copyright (C) 2012, 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Αριθμός Εγγραφής Φ.Π.Α. (Cypriot VAT number).

The Cypriot Αριθμός Εγγραφής Φ.Π.Α. (VAT) number consists of 9 digits
where the last one is a is a letter and functions as a check digit.

>>> compact('CY-10259033P')
'10259033P'
>>> validate('CY-10259033P ')
'10259033P'
>>> validate('CY-10259033Z')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -').upper().strip()
    if number.startswith('CY'):
        number = number[2:]
    return number


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have the
    check digit included."""
    translation = {
        '0': 1, '1': 0, '2': 5, '3': 7, '4': 9,
        '5': 13, '6': 15, '7': 17, '8': 19, '9': 21,
    }
    return 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'[(
        sum(translation[x] for x in number[::2]) +
        sum(int(x) for x in number[1::2])
    ) % 26]


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number[:-1]):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if number[0:2] == '12':
        raise InvalidComponent()
    if number[-1] != calc_check_digit(number[:-1]):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# -*- coding: utf-8 -*-

import logging
from actions.config import get_pre_account_allow_printing
from xml.etree import cElementTree as eTree

import pyscripts
import reportactions
import sysactions

mbcontext = pyscripts.mbcontext

logger = logging.getLogger("PosActions")

auto_print_table_report_cfg = None


class TableEvents(object):
    TOTALED       = "TABLE_TOTALED"
    STORED        = "TABLE_STORED"
    REOPENED      = "TABLE_REOPENED"
    CLOSED        = "TABLE_CLOSED"
    READY4SERVICE = "TABLE_READY4SERVICE"


def main():
    pyscripts.subscribe_event_listener(TableEvents.TOTALED, _event_table_modifier_handler)
    pyscripts.subscribe_event_listener(TableEvents.STORED, _event_table_modifier_handler)
    pyscripts.subscribe_event_listener(TableEvents.REOPENED, _event_table_modifier_handler)
    pyscripts.subscribe_event_listener(TableEvents.CLOSED, _event_table_modifier_handler)
    pyscripts.subscribe_event_listener(TableEvents.READY4SERVICE, _event_table_modifier_handler)


def _event_table_modifier_handler(params):
    event_xml = params[0]
    subject = params[1]
    pos_id = params[4]

    try:
        table_id = eTree.XML(event_xml).find(".//Table").get("id")

        if subject == TableEvents.TOTALED:
            _table_totaled_event(pos_id, table_id)
        elif subject == TableEvents.STORED:
            _table_stored_event(pos_id, table_id)
        elif subject == TableEvents.REOPENED:
            _table_reopened_event(pos_id, table_id)
        elif subject == TableEvents.CLOSED:
            _table_closed_event(pos_id, table_id)
        elif subject == TableEvents.READY4SERVICE:
            _table_ready4service_event(pos_id, table_id)
        else:
            logger.info("[tablelistener] Untreated subject received: {}".format(subject))
    except Exception as _:
        msg = "[tablelistener] Error treating subject: {}".format(subject)
        logger.exception(msg)
        sysactions.sys_log_exception(msg)


def _table_totaled_event(pos_id, table_id):
    _print_table_report(pos_id, table_id, TableEvents.TOTALED)


def _table_stored_event(pos_id, table_id):
    _print_table_report(pos_id, table_id, TableEvents.STORED)


def _table_reopened_event(pos_id, table_id):
    return


def _table_closed_event(pos_id, table_id):
    return


def _table_ready4service_event(pos_id, table_id):
    return


def _print_table_report(pos_id, table_id, table_event):
    auto_print_table_report_events = _get_auto_print_table_report_config()
    if table_event in auto_print_table_report_events:
        model = sysactions.get_model(pos_id)
        order = model.find("CurrentOrder/Order")

        reportactions.print_report(
            pos_id=pos_id,
            report_name="table_report",
            preview=False,
            order=order,
            allow_printing=get_pre_account_allow_printing(),
            report_args=[
                table_id,
                None,
                "TABLE",
                True,
                False,
                0,
            ],
        )


def _get_auto_print_table_report_config():
    global auto_print_table_report_cfg

    if auto_print_table_report_cfg is not None:
        return auto_print_table_report_cfg

    auto_print_table_report = sysactions.get_storewide_config("Store.AutoPrintTableReport")
    auto_print_table_report_cfg = auto_print_table_report.split("\0") if auto_print_table_report else []
    return auto_print_table_report_cfg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Metadata about languages used by our model training code for our
SingleByteCharSetProbers.  Could be used for other things in the future.

This code is based on the language metadata from the uchardet project.
"""
from __future__ import absolute_import, print_function

from string import ascii_letters


# TODO: Add Ukranian (KOI8-U)

class Language(object):
    """Metadata about a language useful for training models

    :ivar name: The human name for the language, in English.
    :type name: str
    :ivar iso_code: 2-letter ISO 639-1 if possible, 3-letter ISO code otherwise,
                    or use another catalog as a last resort.
    :type iso_code: str
    :ivar use_ascii: Whether or not ASCII letters should be included in trained
                     models.
    :type use_ascii: bool
    :ivar charsets: The charsets we want to support and create data for.
    :type charsets: list of str
    :ivar alphabet: The characters in the language's alphabet. If `use_ascii` is
                    `True`, you only need to add those not in the ASCII set.
    :type alphabet: str
    :ivar wiki_start_pages: The Wikipedia pages to start from if we're crawling
                            Wikipedia for training data.
    :type wiki_start_pages: list of str
    """
    def __init__(self, name=None, iso_code=None, use_ascii=True, charsets=None,
                 alphabet=None, wiki_start_pages=None):
        super(Language, self).__init__()
        self.name = name
        self.iso_code = iso_code
        self.use_ascii = use_ascii
        self.charsets = charsets
        if self.use_ascii:
            if alphabet:
                alphabet += ascii_letters
            else:
                alphabet = ascii_letters
        elif not alphabet:
            raise ValueError('Must supply alphabet if use_ascii is False')
        self.alphabet = ''.join(sorted(set(alphabet))) if alphabet else None
        self.wiki_start_pages = wiki_start_pages

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ', '.join('{}={!r}'.format(k, v)
                                         for k, v in self.__dict__.items()
                                         if not k.startswith('_')))


LANGUAGES = {'Arabic': Language(name='Arabic',
                                iso_code='ar',
                                use_ascii=False,
                                # We only support encodings that use isolated
                                # forms, because the current recommendation is
                                # that the rendering system handles presentation
                                # forms. This means we purposefully skip IBM864.
                                charsets=['ISO-8859-6', 'WINDOWS-1256',
                                          'CP720', 'CP864'],
                                alphabet=u'ءآأؤإئابةتثجحخدذرزسشصضطظعغػؼؽؾؿـفقكلمنهوىيًٌٍَُِّ',
                                wiki_start_pages=[u'الصفحة_الرئيسية']),
             'Belarusian': Language(name='Belarusian',
                                    iso_code='be',
                                    use_ascii=False,
                                    charsets=['ISO-8859-5', 'WINDOWS-1251',
                                              'IBM866', 'MacCyrillic'],
                                    alphabet=(u'АБВГДЕЁЖЗІЙКЛМНОПРСТУЎФХЦЧШЫЬЭЮЯ'
                                              u'абвгдеёжзійклмнопрстуўфхцчшыьэюяʼ'),
                                    wiki_start_pages=[u'Галоўная_старонка']),
             'Bulgarian': Language(name='Bulgarian',
                                   iso_code='bg',
                                   use_ascii=False,
                                   charsets=['ISO-8859-5', 'WINDOWS-1251',
                                             'IBM855'],
                                   alphabet=(u'АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЬЮЯ'
                                             u'абвгдежзийклмнопрстуфхцчшщъьюя'),
                                   wiki_start_pages=[u'Начална_страница']),
             'Czech': Language(name='Czech',
                               iso_code='cz',
                               use_ascii=True,
                               charsets=['ISO-8859-2', 'WINDOWS-1250'],
                               alphabet=u'áčďéěíňóřšťúůýžÁČĎÉĚÍŇÓŘŠŤÚŮÝŽ',
                               wiki_start_pages=[u'Hlavní_strana']),
             'Danish': Language(name='Danish',
                                iso_code='da',
                                use_ascii=True,
                                charsets=['ISO-8859-1', 'ISO-8859-15',
                                          'WINDOWS-1252'],
                                alphabet=u'æøåÆØÅ',
                                wiki_start_pages=[u'Forside']),
             'German': Language(name='German',
                                iso_code='de',
                                use_ascii=True,
                                charsets=['ISO-8859-1', 'WINDOWS-1252'],
                                alphabet=u'äöüßÄÖÜ',
                                wiki_start_pages=[u'Wikipedia:Hauptseite']),
             'Greek': Language(name='Greek',
                               iso_code='el',
                               use_ascii=False,
                               charsets=['ISO-8859-7', 'WINDOWS-1253'],
                               alphabet=(u'αβγδεζηθικλμνξοπρσςτυφχψωάέήίόύώ'
                                         u'ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΣΤΥΦΧΨΩΆΈΉΊΌΎΏ'),
                               wiki_start_pages=[u'Πύλη:Κύρια']),
             'English': Language(name='English',
                                 iso_code='en',
                                 use_ascii=True,
                                 charsets=['ISO-8859-1', 'WINDOWS-1252'],
                                 wiki_start_pages=[u'Main_Page']),
             'Esperanto': Language(name='Esperanto',
                                   iso_code='eo',
                                   # Q, W, X, and Y not used at all
                                   use_ascii=False,
                                   charsets=['ISO-8859-3'],
                                   alphabet=(u'abcĉdefgĝhĥijĵklmnoprsŝtuŭvz'
                                             u'ABCĈDEFGĜHĤIJĴKLMNOPRSŜTUŬVZ'),
                                   wiki_start_pages=[u'Vikipedio:Ĉefpaĝo']),
             'Spanish': Language(name='Spanish',
                                 iso_code='es',
                                 use_ascii=True,
                                 charsets=['ISO-8859-1', 'ISO-8859-15',
                                           'WINDOWS-1252'],
                                 alphabet=u'ñáéíóúüÑÁÉÍÓÚÜ',
                                 wiki_start_pages=[u'Wikipedia:Portada']),
             'Estonian': Language(name='Estonian',
                                  iso_code='et',
                                  use_ascii=False,
                                  charsets=['ISO-8859-4', 'ISO-8859-13',
                                            'WINDOWS-1257'],
                                  # C, F, Š, Q, W, X, Y, Z, Ž are only for
                                  # loanwords
                                  alphabet=(u'ABDEGHIJKLMNOPRSTUVÕÄÖÜ'
                                            u'abdeghijklmnoprstuvõäöü'),
                                  wiki_start_pages=[u'Esileht']),
             'Finnish': Language(name='Finnish',
                                 iso_code='fi',
                                 use_ascii=True,
                                 charsets=['ISO-8859-1', 'ISO-8859-15',
                                           'WINDOWS-1252'],
                                 alphabet=u'ÅÄÖŠŽåäöšž',
                                 wiki_start_pages=[u'Wikipedia:Etusivu']),
             'French': Language(name='French',
                                iso_code='fr',
                                use_ascii=True,
                                charsets=['ISO-8859-1', 'ISO-8859-15',
                                          'WINDOWS-1252'],
                                alphabet=u'œàâçèéîïùûêŒÀÂÇÈÉÎÏÙÛÊ',
                                wiki_start_pages=[u'Wikipédia:Accueil_principal',
                                                  u'Bœuf (animal)']),
             'Hebrew': Language(name='Hebrew',
                                iso_code='he',
                                use_ascii=False,
                                charsets=['ISO-8859-8', 'WINDOWS-1255'],
                                alphabet=u'אבגדהוזחטיךכלםמןנסעףפץצקרשתװױײ',
                                wiki_start_pages=[u'עמוד_ראשי']),
             'Croatian': Language(name='Croatian',
                                  iso_code='hr',
                                  # Q, W, X, Y are only used for foreign words.
                                  use_ascii=False,
                                  charsets=['ISO-8859-2', 'WINDOWS-1250'],
                                  alphabet=(u'abcčćdđefghijklmnoprsštuvzž'
                                            u'ABCČĆDĐEFGHIJKLMNOPRSŠTUVZŽ'),
                                  wiki_start_pages=[u'Glavna_stranica']),
             'Hungarian': Language(name='Hungarian',
                                   iso_code='hu',
                                   # Q, W, X, Y are only used for foreign words.
                                   use_ascii=False,
                                   charsets=['ISO-8859-2', 'WINDOWS-1250'],
                                   alphabet=(u'abcdefghijklmnoprstuvzáéíóöőúüű'
                                             u'ABCDEFGHIJKLMNOPRSTUVZÁÉÍÓÖŐÚÜŰ'),
                                   wiki_start_pages=[u'Kezdőlap']),
             'Italian': Language(name='Italian',
                                 iso_code='it',
                                 use_ascii=True,
                                 charsets=['ISO-8859-1', 'ISO-8859-15',
                                           'WINDOWS-1252'],
                                 alphabet=u'ÀÈÉÌÒÓÙàèéìòóù',
                                 wiki_start_pages=[u'Pagina_principale']),
             'Lithuanian': Language(name='Lithuanian',
                                    iso_code='lt',
                                    use_ascii=False,
                                    charsets=['ISO-8859-13', 'WINDOWS-1257',
                                              'ISO-8859-4'],
                                    # Q, W, and X not used at all
                                    alphabet=(u'AĄBCČDEĘĖFGHIĮYJKLMNOPRSŠTUŲŪVZŽ'
                                              u'aąbcčdeęėfghiįyjklmnoprsštuųūvzž'),
                                    wiki_start_pages=[u'Pagrindinis_puslapis']),
             'Latvian': Language(name='Latvian',
                                 iso_code='lv',
                                 use_ascii=False,
                                 charsets=['ISO-8859-13', 'WINDOWS-1257',
                                           'ISO-8859-4'],
                                 # Q, W, X, Y are only for loanwords
                                 alphabet=(u'AĀBCČDEĒFGĢHIĪJKĶLĻMNŅOPRSŠTUŪVZŽ'
                                           u'aābcčdeēfgģhiījkķlļmnņoprsštuūvzž'),
                                 wiki_start_pages=[u'Sākumlapa']),
             'Macedonian': Language(name='Macedonian',
                                    iso_code='mk',
                                    use_ascii=False,
                                    charsets=['ISO-8859-5', 'WINDOWS-1251',
                                              'MacCyrillic', 'IBM855'],
                                    alphabet=(u'АБВГДЃЕЖЗЅИЈКЛЉМНЊОПРСТЌУФХЦЧЏШ'
                                              u'абвгдѓежзѕијклљмнњопрстќуфхцчџш'),
                                    wiki_start_pages=[u'Главна_страница']),
             'Dutch': Language(name='Dutch',
                               iso_code='nl',
                               use_ascii=True,
                               charsets=['ISO-8859-1', 'WINDOWS-1252'],
                               wiki_start_pages=[u'Hoofdpagina']),
             'Polish': Language(name='Polish',
                                iso_code='pl',
                                # Q and X are only used for foreign words.
                                use_ascii=False,
                                charsets=['ISO-8859-2', 'WINDOWS-1250'],
                                alphabet=(u'AĄBCĆDEĘFGHIJKLŁMNŃOÓPRSŚTUWYZŹŻ'
                                          u'aąbcćdeęfghijklłmnńoóprsśtuwyzźż'),
                                wiki_start_pages=[u'Wikipedia:Strona_główna']),
             'Portuguese': Language(name='Portuguese',
                                 iso_code='pt',
                                 use_ascii=True,
                                 charsets=['ISO-8859-1', 'ISO-8859-15',
                                           'WINDOWS-1252'],
                                 alphabet=u'ÁÂÃÀÇÉÊÍÓÔÕÚáâãàçéêíóôõú',
                                 wiki_start_pages=[u'Wikipédia:Página_principal']),
             'Romanian': Language(name='Romanian',
                                  iso_code='ro',
                                  use_ascii=True,
                                  charsets=['ISO-8859-2', 'WINDOWS-1250'],
                                  alphabet=u'ăâîșțĂÂÎȘȚ',
                                  wiki_start_pages=[u'Pagina_principală']),
             'Russian': Language(name='Russian',
                                 iso_code='ru',
                                 use_ascii=False,
                                 charsets=['ISO-8859-5', 'WINDOWS-1251',
                                           'KOI8-R', 'MacCyrillic', 'IBM866',
                                           'IBM855'],
                                 alphabet=(u'абвгдеёжзийклмнопрстуфхцчшщъыьэюя'
                                           u'АБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ'),
                                 wiki_start_pages=[u'Заглавная_страница']),
             'Slovak': Language(name='Slovak',
                                iso_code='sk',
                                use_ascii=True,
                                charsets=['ISO-8859-2', 'WINDOWS-1250'],
                                alphabet=u'áäčďéíĺľňóôŕšťúýžÁÄČĎÉÍĹĽŇÓÔŔŠŤÚÝŽ',
                                wiki_start_pages=[u'Hlavná_stránka']),
             'Slovene': Language(name='Slovene',
                                 iso_code='sl',
                                 # Q, W, X, Y are only used for foreign words.
                                 use_ascii=False,
                                 charsets=['ISO-8859-2', 'WINDOWS-1250'],
                                 alphabet=(u'abcčdefghijklmnoprsštuvzž'
                                           u'ABCČDEFGHIJKLMNOPRSŠTUVZŽ'),
                                 wiki_start_pages=[u'Glavna_stran']),
             # Serbian can be written in both Latin and Cyrillic, but there's no
             # simple way to get the Latin alphabet pages from Wikipedia through
             # the API, so for now we just support Cyrillic.
             'Serbian': Language(name='Serbian',
                                 iso_code='sr',
                                 alphabet=(u'АБВГДЂЕЖЗИЈКЛЉМНЊОПРСТЋУФХЦЧЏШ'
                                           u'абвгдђежзијклљмнњопрстћуфхцчџш'),
                                 charsets=['ISO-8859-5', 'WINDOWS-1251',
                                           'MacCyrillic', 'IBM855'],
                                 wiki_start_pages=[u'Главна_страна']),
             'Thai': Language(name='Thai',
                              iso_code='th',
                              use_ascii=False,
                              charsets=['ISO-8859-11', 'TIS-620', 'CP874'],
                              alphabet=u'กขฃคฅฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืฺุู฿เแโใไๅๆ็่้๊๋์ํ๎๏๐๑๒๓๔๕๖๗๘๙๚๛',
                              wiki_start_pages=[u'หน้าหลัก']),
             'Turkish': Language(name='Turkish',
                                 iso_code='tr',
                                 # Q, W, and X are not used by Turkish
                                 use_ascii=False,
                                 charsets=['ISO-8859-3', 'ISO-8859-9',
                                           'WINDOWS-1254'],
                                 alphabet=(u'abcçdefgğhıijklmnoöprsştuüvyzâîû'
                                           u'ABCÇDEFGĞHIİJKLMNOÖPRSŞTUÜVYZÂÎÛ'),
                                 wiki_start_pages=[u'Ana_Sayfa']),
             'Vietnamese': Language(name='Vietnamese',
                                    iso_code='vi',
                                    use_ascii=False,
                                    # Windows-1258 is the only common 8-bit
                                    # Vietnamese encoding supported by Python.
                                    # From Wikipedia:
                                    # For systems that lack support for Unicode,
                                    # dozens of 8-bit Vietnamese code pages are
                                    # available.[1] The most common are VISCII
                                    # (TCVN 5712:1993), VPS, and Windows-1258.[3]
                                    # Where ASCII is required, such as when
                                    # ensuring readability in plain text e-mail,
                                    # Vietnamese letters are often encoded
                                    # according to Vietnamese Quoted-Readable
                                    # (VIQR) or VSCII Mnemonic (VSCII-MNEM),[4]
                                    # though usage of either variable-width
                                    # scheme has declined dramatically following
                                    # the adoption of Unicode on the World Wide
                                    # Web.
                                    charsets=['WINDOWS-1258'],
                                    alphabet=(u'aăâbcdđeêghiklmnoôơpqrstuưvxy'
                                              u'AĂÂBCDĐEÊGHIKLMNOÔƠPQRSTUƯVXY'),
                                    wiki_start_pages=[u'Chữ_Quốc_ngữ']),
            }



# cups.py - functions for handling Spanish CUPS code
# coding: utf-8
#
# Copyright (C) 2016 David García Garzón
# Copyright (C) 2016 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CUPS (Código Unificado de Punto de Suministro, Spanish meter point number).

CUPS codes are used in Spain as unique identifier for energy supply points.
They are used both for electricity and pipelined gas.

The format is set by the Energy Ministry, and individual codes are issued by
each local distribution company. The number consist or 20 or 22 digits and is
built up as follows:

* LL: (letters) country (always 'ES' since it is a national code)
* DDDD: (numbers) distribution company code (numeric)
* CCCC CCCC CCCC: identifier within the distribution company (numeric)
* EE: (letters) check digits
* N: (number) border point sequence
* T: (letter) kind of border point

More information:

* https://es.wikipedia.org/wiki/Código_Unificado_de_Punto_de_Suministro

>>> validate('ES 1234-123456789012-JY')
'ES1234123456789012JY'
>>> validate('ES 1234-123456789012-JY 1T')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('ES 1234-123456789012-XY 1F')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('ES1234123456789012JY1F')
'ES 1234 1234 5678 9012 JY 1F'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip().upper()


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((
        number[:2],
        number[2:6],
        number[6:10],
        number[10:14],
        number[14:18],
        number[18:20],
        number[20:],
    )).strip()


def calc_check_digits(number):
    """Calculate the check digits for the number."""
    alphabet = 'TRWAGMYFPDXBNJZSQVHLCKE'
    check0, check1 = divmod(int(number[2:18]) % 529, 23)
    return alphabet[check0] + alphabet[check1]


def validate(number):
    """Check if the number provided is a valid CUPS. This checks length,
    formatting and check digits."""
    number = compact(number)
    if len(number) not in (20, 22):
        raise InvalidLength()
    if number[:2] != 'ES':
        raise InvalidComponent()
    if not isdigits(number[2:18]):
        raise InvalidFormat()
    if number[20:]:
        pnumber, ptype = number[20:]
        if not isdigits(pnumber):
            raise InvalidFormat()
        if ptype not in 'FPRCXYZ':
            raise InvalidFormat()
    if calc_check_digits(number) != number[18:20]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid CUPS."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



def get_operator_info(model, operator_id):
    need_logged = True
    session = ''
    for op in model.find('Operators').findall('Operator'):
        if op.get('id') == operator_id:
            operator = op
            session = operator.get('sessionId')
            operator_id = operator.get("id")
            if op.get('state') == 'PAUSED':
                need_logged = False
            break

    return need_logged, operator_id, session



from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder
from production.box import TagItemsWhenProdStateBox


class TagItemsWhenProdStateBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(TagItemsWhenProdStateBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> TagItemsWhenProdStateBox
        return TagItemsWhenProdStateBox(config.name,
                                        config.sons,
                                        self.get_extra(config, "ProdState", ""),
                                        self.get_extra(config, "Tag", ""),
                                        self.get_logger(config))



from emptysearchexception import EmptySearchException
from voidorderexception import VoidOrderException
from ._FiscalException import (
    FiscalIncrementException,
    FiscalGetIdException,
    FiscalParametersException,
    FiscalValidation,
)
from ._StoreNotOperatorWithSatException import StoreNotOperatorWithSatException



from application.interactor import GetCatalogVersionToApplyInteractor
from messagebus import Message, DefaultToken, DataType
from messageprocessorutil import UuidMessageProcessor
from application.util import settings


class GetCatalogVersionToApplyProcessor(UuidMessageProcessor):

    def __init__(self, interactor, event_name=None):
        # type: (GetCatalogVersionToApplyInteractor, str) -> None
        super(GetCatalogVersionToApplyProcessor, self).__init__(event_name)
        self.interactor = interactor

    def parse_data(self, data, event_type):
        return data.lower() == "true"

    def call_business(self, need_backoffice_json):

        settings.clear_status()
        return self.interactor.get_catalog_version_to_apply(need_backoffice_json)

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.param, result))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        exception_name = str(type(exception).__name__)
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, exception_name))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        return



# -*- coding: utf-8 -*-

import logging
from abc import abstractmethod, ABCMeta

from old_helper import round_half_away_from_zero

from comp_exceptions import FiscalParametersException

logger = logging.getLogger("FiscalWrapper")


class TaxBuilder(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def calculate_icms(self, cst_icms, csosn, tax_item, fcp_item, base_reduction, reason, product_code):
        pass

    @abstractmethod
    def calculate_pis(self, cst_pis, tax_item):
        pass

    @abstractmethod
    def calculate_cofins(self, cst_cofins, tax_item):
        pass


class TaxBuilderLucroReal(TaxBuilder):
    def __init__(self):
        super(TaxBuilder, self).__init__()

    ICMS_00 = "<ICMS>" \
                "<ICMS00>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "<modBC>{mod_bc}</modBC>" \
                  "<vBC>{v_bc:.02f}</vBC>" \
                  "<pICMS>{p_icms:.04f}</pICMS>" \
                  "<vICMS>{v_icms:.02f}</vICMS>" \
                  "{fcp_data}" \
                "</ICMS00>" \
              "</ICMS>"

    ICMS_00_FCP = "<pFCP>{p_fcp:.04f}</pFCP>" \
                  "<vFCP>{v_fcp:.02f}</vFCP>"

    ICMS_20 = "<ICMS>" \
                "<ICMS20>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "<modBC>{mod_bc}</modBC>" \
                  "<pRedBC>{p_red_bc:.02f}</pRedBC>" \
                  "<vBC>{v_bc:.02f}</vBC>" \
                  "<pICMS>{p_icms:.04f}</pICMS>" \
                  "<vICMS>{v_icms:.02f}</vICMS>" \
                  "{fcp_data}" \
                  "{icms_deson_data}" \
              "</ICMS20>" \
              "</ICMS>"

    ICMS_20_FCP = "<vBCFCP>{bc_fcp:.02f}</vBCFCP>" \
                  "<pFCP>{p_fcp:.04f}</pFCP>" \
                  "<vFCP>{v_fcp:.02f}</vFCP>"

    ICMS_40 = "<ICMS>" \
                "<ICMS40>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "{icms_deson_data}" \
                "</ICMS40>" \
              "</ICMS>"

    ICMS_DESON = "<vICMSDeson>{deson:.02f}</vICMSDeson>" \
                     "<motDesICMS>{reason}</motDesICMS>"

    ICMS_41 = "<ICMS>" \
                "<ICMS40>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                "</ICMS40>" \
              "</ICMS>"

    ICMS_60 = "<ICMS>" \
                "<ICMS60>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "{fcp_data}" \
                  "<pRedBCEfet>0.00</pRedBCEfet>" \
                  "<vBCEfet>0.00</vBCEfet>" \
                  "<pICMSEfet>0.0000</pICMSEfet>" \
                  "<vICMSEfet>0.00</vICMSEfet>" \
              "</ICMS60>" \
              "</ICMS>"

    ICMS_60_FCP = "<vBCFCPSTRet>{bc_fcp:.02f}</vBCFCPSTRet>" \
                  "<pFCPSTRet>{p_fcp:.04f}</pFCPSTRet>" \
                  "<vFCPSTRet>{v_fcp:.02f}</vFCPSTRet>"

    PIS_NT = "<PIS>" \
               "<PISNT>" \
                 "<CST>{cst:>02}</CST>" \
               "</PISNT>" \
             "</PIS>"

    PIS_ALIQ = "<PIS>" \
                 "<PISAliq>" \
                   "<CST>{cst:>02}</CST>" \
                   "<vBC>{v_bc:.02f}</vBC>" \
                   "<pPIS>{p_pis:.04f}</pPIS>" \
                   "<vPIS>{v_pis:0.2f}</vPIS>" \
                 "</PISAliq>" \
               "</PIS>"

    COFINS_NT = "<COFINS>" \
                  "<COFINSNT>" \
                    "<CST>{cst:>02}</CST>" \
                  "</COFINSNT>" \
                "</COFINS>"

    COFINS_ALIQ = "<COFINS>" \
                    "<COFINSAliq>" \
                      "<CST>{cst:>02}</CST>" \
                      "<vBC>{v_bc:.02f}</vBC>" \
                      "<pCOFINS>{p_cofins:.04f}</pCOFINS>" \
                      "<vCOFINS>{v_cofins:0.2f}</vCOFINS>" \
                    "</COFINSAliq>" \
                  "</COFINS>"

    PIS_OUTR = "<PIS>" \
               "<PISOutr>" \
               "<CST>{cst:0>2}</CST>" \
               "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
               "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
               "<vPIS>{v_pis:0.2f}</vPIS>" \
               "</PISOutr>" \
               "</PIS>"

    COFINS_OUTR = "<COFINS>" \
                  "<COFINSOutr>" \
                  "<CST>{cst:0>2}</CST>" \
                  "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
                  "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
                  "<vCOFINS>{v_cofins:0.2f}</vCOFINS>" \
                  "</COFINSOutr>" \
                  "</COFINS>"

    def calculate_icms(self, cst_icms, csosn, tax_item, fcp_item, base_reduction, reason, product_code):
        # returns xml line, base amount, tax amount and deson
        _round_tax_amount(tax_item)
        if cst_icms == 0 and tax_item:
            fcp_data = ""
            fcp_tax = 0
            if fcp_item is not None:
                fcp_data = self.ICMS_00_FCP.format(p_fcp=fcp_item.tax_rate*100,
                                                   v_fcp=fcp_item.tax_amount_ad)
                fcp_tax = fcp_item.tax_amount_ad
            xml = self.ICMS_00.format(orig=0,
                                      cst=cst_icms,
                                      mod_bc=3,
                                      v_bc=tax_item.base_amount_ad,
                                      p_icms=tax_item.tax_rate * 100,
                                      v_icms=tax_item.tax_amount_ad,
                                      fcp_data=fcp_data)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad, 0, fcp_tax, 0, 0

        elif cst_icms == 20 and tax_item:
            fcp_data = ""
            fcp_tax = 0
            if fcp_item is not None:
                fcp_data = self.ICMS_20_FCP.format(
                    bc_fcp=tax_item.base_amount_ad,
                    p_fcp=fcp_item.tax_rate * 100,
                    v_fcp=fcp_item.tax_amount_ad
                )
                fcp_tax = fcp_item.tax_amount_ad

            icms_deson = 0
            icms_deson_data = ''
            if base_reduction > 0:
                if not reason:
                    reason = "9"

                base_amount_without_reduction = tax_item.base_amount_ad / (1 - (base_reduction / 100))
                tax_amount_without_reduction = base_amount_without_reduction * tax_item.tax_rate
                icms_deson = tax_amount_without_reduction - tax_item.tax_amount_ad
                icms_deson_data = self.ICMS_DESON.format(deson=icms_deson, reason=reason)

            xml = self.ICMS_20.format(
                orig=0,
                cst=cst_icms,
                mod_bc=3,
                v_bc=tax_item.base_amount_ad,
                p_red_bc=base_reduction,
                p_icms=tax_item.tax_rate * 100,
                v_icms=tax_item.tax_amount_ad,
                fcp_data=fcp_data,
                icms_deson_data=icms_deson_data
            )
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad, icms_deson, fcp_tax, 0, 0

        elif cst_icms == 40:
            amount_icms = 0
            if tax_item:
                amount_icms = tax_item.tax_amount_ad

            if amount_icms > 0:
                logger.warning("CST ICMS = 40, tax amount maior que zero")

            if not reason:
                logger.warning("CST ICMS = 40, não possui motivo, usando padrão 9")
                reason = "9"

            xml = self.ICMS_40.format(
                orig=0,
                cst=cst_icms,
                icms_deson_data=self.ICMS_DESON.format(deson=amount_icms, reason=reason) if reason else ''
            )

            return xml, 0, 0, amount_icms if reason else 0, 0, 0, 0

        elif cst_icms == 41:
            xml = self.ICMS_41.format(
                orig=0,
                cst=cst_icms,
            )

            return xml, 0, 0, 0, 0, 0, 0

        elif cst_icms == 60 and tax_item:
            fcp_data = ""
            fcp_tax = 0
            fcp_tax_rate = 0
            if fcp_item is not None:
                fcp_data = self.ICMS_60_FCP.format(
                    bc_fcp=tax_item.base_amount_ad,
                    p_fcp=fcp_item.tax_rate * 100,
                    v_fcp=fcp_item.tax_amount_ad
                )

                fcp_tax = fcp_item.tax_amount_ad
                fcp_tax_rate = fcp_item.tax_rate * 100
            if tax_item.tax_amount_ad > 0:
                logger.warning("CST ICMS = 60, tax amount maior que zero")

            xml = self.ICMS_60.format(orig=0, cst=cst_icms, pst=tax_item.tax_rate*100 + fcp_tax_rate, fcp_data=fcp_data)
            return xml, 0, 0, 0, 0, 0, fcp_tax

        else:
            if not tax_item:
                raise FiscalParametersException("Sem ICMS para produto {}".format(product_code))
            raise FiscalParametersException("CST ICMS/Taxa Desconhecido {} Produto {}".format(cst_icms, product_code))

    def calculate_pis(self, cst_pis, tax_item):
        _round_tax_amount(tax_item)
        if cst_pis in (1, 2, 99) and tax_item:
            xml = self.PIS_ALIQ.format(cst=cst_pis,
                                       v_bc=tax_item.base_amount_ad,
                                       p_pis=tax_item.tax_rate * 100,
                                       v_pis=tax_item.tax_amount_ad)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad

        elif cst_pis == 49 and tax_item:
            xml = self.PIS_OUTR.format(cst=cst_pis,
                                       q_bc_prod=tax_item.base_amount_ad,
                                       v_aliq_prod=tax_item.tax_amount_ad,
                                       v_pis=tax_item.tax_amount_ad)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad

        elif cst_pis == 99:
            xml = self.PIS_OUTR.format(cst=cst_pis,
                                       q_bc_prod=0,
                                       v_aliq_prod=0,
                                       v_pis=0)
            return xml, 0, 0

        elif cst_pis in (4, 5, 6, 7, 8, 9):
            return self.PIS_NT.format(cst=cst_pis), 0, 0

        else:
            raise FiscalParametersException("CST PIS/Taxa Desconhecido %s" % cst_pis)

    def calculate_cofins(self, cst_cofins, tax_item):
        _round_tax_amount(tax_item)
        if cst_cofins in (1, 2, 99) and tax_item:
            xml = self.COFINS_ALIQ.format(cst=cst_cofins,
                                          v_bc=tax_item.base_amount_ad,
                                          p_cofins=tax_item.tax_rate * 100,
                                          v_cofins=tax_item.tax_amount_ad)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad

        elif cst_cofins == 49 and tax_item:
            xml = self.COFINS_OUTR.format(cst=cst_cofins,
                                          q_bc_prod=tax_item.base_amount_ad,
                                          v_aliq_prod=tax_item.tax_amount_ad,
                                          v_cofins=tax_item.tax_amount_ad)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad

        elif cst_cofins == 99:
            xml = self.COFINS_OUTR.format(cst=cst_cofins,
                                          q_bc_prod=0,
                                          v_aliq_prod=0,
                                          v_cofins=0)
            return xml, 0, 0

        elif cst_cofins in (4, 5, 6, 7, 8, 9):
            return self.COFINS_NT.format(cst=cst_cofins), 0, 0

        else:
            raise FiscalParametersException("CST COFINS/Taxa Desconhecido %s" % cst_cofins)


class TaxBuilderSimplesNacional(TaxBuilder):
    def __init__(self):
        super(TaxBuilderSimplesNacional, self).__init__()

    ICMSSN_101 = "<ICMS>" \
                   "<ICMSSN101>" \
                     "<orig>{orig}</orig>" \
                     "<CSOSN>{csosn}</CSOSN>" \
                   "</ICMSSN101>" \
                 "</ICMS>"

    ICMSSN_102 = "<ICMS>" \
                   "<ICMSSN102>" \
                     "<orig>{orig}</orig>" \
                     "<CSOSN>{csosn}</CSOSN>" \
                   "</ICMSSN102>" \
                 "</ICMS>"

    ICMSSN_201 = "<ICMS>" \
                   "<ICMSSN201>" \
                     "<orig>{orig}</orig>" \
                     "<CSOSN>{csosn}</CSOSN>" \
                   "</ICMSSN201>" \
                 "</ICMS>"

    ICMSSN_202 = "<ICMS>" \
                   "<ICMSSN202>" \
                     "<orig>{orig}</orig>" \
                     "<CSOSN>{csosn}</CSOSN>" \
                   "</ICMSSN202>" \
                 "</ICMS>"

    ICMSSN_500 = "<ICMS>" \
                   "<ICMSSN500>" \
                     "<orig>{orig}</orig>" \
                     "<CSOSN>{csosn}</CSOSN>" \
                     "{fcp_data}" \
                     "<pRedBCEfet>0.00</pRedBCEfet>" \
                     "<vBCEfet>0.00</vBCEfet>" \
                     "<pICMSEfet>0.0000</pICMSEfet>" \
                     "<vICMSEfet>0.00</vICMSEfet>" \
                 "</ICMSSN500>" \
                 "</ICMS>"

    ICMSSN_500_FCP = "<vBCFCPSTRet>{bc_fcp:.02f}</vBCFCPSTRet>" \
                     "<pFCPSTRet>{p_fcp:.04f}</pFCPSTRet>" \
                     "<vFCPSTRet>{v_fcp:.02f}</vFCPSTRet>"

    ICMSSN_900 = "<ICMS>" \
                     "<ICMSSN900>" \
                         "<orig>{orig}</orig>" \
                         "<CSOSN>{csosn}</CSOSN>" \
                     "</ICMSSN900>" \
                 "</ICMS>"

    PIS_OUTR = "<PIS>" \
                 "<PISOutr>" \
                   "<CST>{cst:0>2}</CST>" \
                   "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
                   "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
                   "<vPIS>{v_pis:0.2f}</vPIS>" \
                 "</PISOutr>" \
               "</PIS>"

    PIS_NT = "<PIS>" \
               "<PISNT>" \
                 "<CST>{cst:>02}</CST>" \
               "</PISNT>" \
             "</PIS>"

    COFINS_OUTR = "<COFINS>" \
                    "<COFINSOutr>" \
                      "<CST>{cst:0>2}</CST>" \
                      "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
                      "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
                      "<vCOFINS>{v_cofins:0.2f}</vCOFINS>" \
                    "</COFINSOutr>" \
                  "</COFINS>"

    COFINS_NT = "<COFINS>" \
                  "<COFINSNT>" \
                    "<CST>{cst:>02}</CST>" \
                  "</COFINSNT>" \
                "</COFINS>"

    def calculate_icms(self, cst_icms, csosn, tax_item, fcp_item, base_reduction, reason, product_code):
        csosn = int(csosn)

        if csosn in [101]:
            return self.ICMSSN_101.format(orig=0, csosn=csosn), 0, 0, 0, 0, 0, 0
        if csosn in [102, 103, 300, 400]:
            return self.ICMSSN_102.format(orig=0, csosn=csosn), 0, 0, 0, 0, 0, 0
        elif csosn in [201]:
            return self.ICMSSN_201.format(orig=0, csosn=csosn), 0, 0, 0, 0, 0, 0
        elif csosn in [202, 203]:
            return self.ICMSSN_202.format(orig=0, csosn=csosn), 0, 0, 0, 0, 0, 0
        elif csosn in [500]:
            fcp_data = ""
            fcp_tax = 0

            if fcp_item and tax_item:
                fcp_tax = fcp_item.tax_amount_ad
                fcp_rate = fcp_item.tax_rate * 100
                fcp_data = self.ICMSSN_500_FCP.format(bc_fcp=tax_item.base_amount_ad, p_fcp=fcp_rate, v_fcp=fcp_tax)

            return self.ICMSSN_500.format(orig=0, csosn=csosn, fcp_data=fcp_data), 0, 0, 0, 0, 0, fcp_tax
        elif csosn in [900]:
            return self.ICMSSN_900.format(orig=0, csosn=csosn), 0, 0, 0, 0, 0, 0
        else:
            raise FiscalParametersException("CSOSN Desconhecido {} Produto {}".format(csosn, product_code))

    def calculate_pis(self, cst_pis, tax_item):
        if cst_pis in (1, 2, 49, 99):
            return self.PIS_OUTR.format(cst=99, q_bc_prod=0, v_aliq_prod=0, v_pis=0), 0, 0
        elif cst_pis in (4, 5, 6, 7, 8, 9):
            return self.PIS_NT.format(cst=cst_pis), 0, 0
        else:
            raise FiscalParametersException("CST PIS Desconhecido %s" % cst_pis)

    def calculate_cofins(self, cst_cofins, tax_item):
        if cst_cofins in (1, 2, 49, 99):
            return self.COFINS_OUTR.format(cst=99, q_bc_prod=0, v_aliq_prod=0, v_cofins=0), 0, 0
        elif cst_cofins in (4, 5, 6, 7, 8, 9):
            return self.COFINS_NT.format(cst=cst_cofins), 0, 0
        else:
            raise FiscalParametersException("CST COFINS Desconhecido %s" % cst_cofins)


class TaxBuilderSimplesNacionalSublimiteReceita(TaxBuilder):
    def __init__(self):
        super(TaxBuilderSimplesNacionalSublimiteReceita, self).__init__()

    ICMS_00 = "<ICMS>" \
                "<ICMS00>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "<modBC>{mod_bc}</modBC>" \
                  "<vBC>{v_bc:.02f}</vBC>" \
                  "<pICMS>{p_icms:.04f}</pICMS>" \
                  "<vICMS>{v_icms:.02f}</vICMS>" \
                  "{fcp_data}" \
                  "</ICMS00>" \
                "</ICMS>"

    ICMS_00_FCP = "<pFCP>{p_fcp:.04f}</pFCP>" \
                  "<vFCP>{v_fcp:.02f}</vFCP>"

    ICMS_20 = "<ICMS>" \
                "<ICMS20>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "<modBC>{mod_bc}</modBC>" \
                  "<pRedBC>{p_red_bc:.02f}</pRedBC>" \
                  "<vBC>{v_bc:.02f}</vBC>" \
                  "<pICMS>{p_icms:.04f}</pICMS>" \
                  "<vICMS>{v_icms:.02f}</vICMS>" \
                  "{fcp_data}" \
                  "{icms_deson_data}" \
              "</ICMS20>" \
              "</ICMS>"

    ICMS_20_FCP = "<vBCFCP>{bc_fcp:.02f}</vBCFCP>" \
                  "<pFCP>{p_fcp:.04f}</pFCP>" \
                  "<vFCP>{v_fcp:.02f}</vFCP>"

    ICMS_40 = "<ICMS>" \
                "<ICMS40>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "{icms_deson_data}" \
              "</ICMS40>" \
              "</ICMS>"

    ICMS_DESON = "<vICMSDeson>{deson:.02f}</vICMSDeson>" \
                     "<motDesICMS>{reason}</motDesICMS>"
    ICMS_60 = "<ICMS>" \
                "<ICMS60>" \
                  "<orig>{orig}</orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "{fcp_data}" \
                  "<pRedBCEfet>0.00</pRedBCEfet>" \
                  "<vBCEfet>0.00</vBCEfet>" \
                  "<pICMSEfet>0.0000</pICMSEfet>" \
                  "<vICMSEfet>0.00</vICMSEfet>" \
              "</ICMS60>" \
              "</ICMS>"

    ICMS_60_FCP = "<vBCFCPSTRet>{bc_fcp:.02f}</vBCFCPSTRet>" \
                  "<pFCPSTRet>{p_fcp:.04f}</pFCPSTRet>" \
                  "<vFCPSTRet>{v_fcp:.02f}</vFCPSTRet>"

    PIS_OUTR = "<PIS>" \
                 "<PISOutr>" \
                   "<CST>{cst:0>2}</CST>" \
                   "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
                   "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
                   "<vPIS>{v_pis:0.2f}</vPIS>" \
                 "</PISOutr>" \
               "</PIS>"

    PIS_NT = "<PIS>" \
               "<PISNT>" \
                 "<CST>{cst:>02}</CST>" \
               "</PISNT>" \
             "</PIS>"

    COFINS_OUTR = "<COFINS>" \
                    "<COFINSOutr>" \
                      "<CST>{cst:0>2}</CST>" \
                      "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
                      "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
                      "<vCOFINS>{v_cofins:0.2f}</vCOFINS>" \
                    "</COFINSOutr>" \
                  "</COFINS>"

    COFINS_NT = "<COFINS>" \
                  "<COFINSNT>" \
                    "<CST>{cst:>02}</CST>" \
                  "</COFINSNT>" \
                "</COFINS>"

    def calculate_icms(self, cst_icms, csosn, tax_item, fcp_item, base_reduction, reason, product_code):
        _round_tax_amount(tax_item)
        if cst_icms in (0, ):
            fcp_data = ""
            fcp_tax = 0
            if fcp_item is not None:
                fcp_data = self.ICMS_00_FCP.format(p_fcp=fcp_item.tax_rate * 100, v_fcp=fcp_item.tax_amount_ad)
                fcp_tax = fcp_item.tax_amount_ad

            return self.ICMS_00.format(orig=0,
                                       cst=0,
                                       mod_bc=3,
                                       v_bc=tax_item.base_amount_ad,
                                       p_icms=tax_item.tax_rate * 100,
                                       v_icms=tax_item.tax_amount_ad,
                                       fcp_data=fcp_data), tax_item.base_amount_ad, tax_item.tax_amount_ad, 0, fcp_tax, 0, 0
        elif cst_icms in (60, ):
            fcp_data = ""
            fcp_tax = 0
            fcp_tax_rate = 0
            if fcp_item is not None:
                fcp_data = self.ICMS_60_FCP.format(bc_fcp=tax_item.base_amount_ad,
                                                   p_fcp=fcp_item.tax_rate * 100,
                                                   v_fcp=fcp_item.tax_amount_ad)
                fcp_tax = fcp_item.tax_amount_ad
                fcp_tax_rate = fcp_item.tax_rate * 100

            return self.ICMS_60.format(orig=0,
                                       cst=cst_icms,
                                       pst=tax_item.tax_rate * 100 + fcp_tax_rate,
                                       fcp_data=fcp_data), 0, 0, 0, 0, 0, fcp_tax
        elif cst_icms in (20, ):
            fcp_data = ""
            fcp_tax = 0
            if fcp_item is not None:
                fcp_data = self.ICMS_20_FCP.format(bc_fcp=tax_item.base_amount_ad,
                                                   p_fcp=fcp_item.tax_rate * 100,
                                                   v_fcp=fcp_item.tax_amount_ad)
                fcp_tax = fcp_item.tax_amount_ad

            icms_deson = 0
            icms_deson_data = ''
            if base_reduction > 0:
                if not reason:
                    reason = "9"

                base_amount_without_reduction = tax_item.base_amount_ad / (1 - (base_reduction / 100))
                tax_amount_without_reduction = base_amount_without_reduction * tax_item.tax_rate
                icms_deson = tax_amount_without_reduction - tax_item.tax_amount_ad
                icms_deson_data = self.ICMS_DESON.format(deson=icms_deson, reason=reason)

            xml = self.ICMS_20.format(
                orig=0,
                cst=cst_icms,
                mod_bc=3,
                v_bc=tax_item.base_amount_ad,
                p_red_bc=base_reduction,
                p_icms=tax_item.tax_rate * 100,
                v_icms=tax_item.tax_amount_ad,
                fcp_data=fcp_data,
                icms_deson_data=icms_deson_data
            )

            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad, icms_deson, fcp_tax, 0, 0
        elif cst_icms in (40, 41, 50, 51):
            return self.ICMS_40.format(orig=0,
                                       cst=cst_icms,
                                       icms_deson_data=self.ICMS_DESON.format(deson=tax_item.tax_amount_ad, reason=reason) if reason else ''),\
                   0, 0, tax_item.tax_amount_ad if reason else 0, 0, 0, 0

        else:
            if not tax_item:
                raise FiscalParametersException("Sem ICMS para produto {}".format(product_code))
            raise FiscalParametersException("CST ICMS/Taxa Desconhecido {} Produto {}".format(cst_icms, product_code))

    def calculate_pis(self, cst_pis, tax_item):
        if cst_pis in (1, 2, 99):
            return self.PIS_OUTR.format(cst=99, q_bc_prod=0, v_aliq_prod=0, v_pis=0), 0, 0
        elif cst_pis in (4, 5, 6, 7, 8, 9):
            return self.PIS_NT.format(cst=cst_pis), 0, 0
        else:
            raise FiscalParametersException("CST PIS/Taxa Desconhecido %s" % cst_pis)

    def calculate_cofins(self, cst_cofins, tax_item):
        if cst_cofins in (1, 2, 99):
            return self.COFINS_OUTR.format(cst=99, q_bc_prod=0, v_aliq_prod=0, v_cofins=0), 0, 0
        elif cst_cofins in (4, 5, 6, 7, 8, 9):
            return self.COFINS_NT.format(cst=cst_cofins), 0, 0
        else:
            raise FiscalParametersException("CST COFINS/Taxa Desconhecido %s" % cst_cofins)


def _round_tax_amount(tax_item):
    if tax_item:
        tax_item.base_amount_ad = round_half_away_from_zero(tax_item.base_amount_ad, 2)
        tax_item.tax_amount_ad = round_half_away_from_zero(tax_item.tax_amount_ad, 2)



# cnp.py - functions for handling Romanian CNP numbers
# coding: utf-8
#
# Copyright (C) 2012-2020 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CNP (Cod Numeric Personal, Romanian Numerical Personal Code).

The CNP is a 13 digit number that includes information on the person's
gender, birth date and country zone.

More information:

* https://ro.wikipedia.org/wiki/Cod_numeric_personal
* https://github.com/vimishor/cnp-spec/blob/master/spec.md

>>> validate('1630615123457')
'1630615123457'
>>> get_county('1630615123457')
'Cluj'
>>> validate('0800101221142')  # invalid first digit
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('1632215123457')  # invalid date
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('1630615993454')  # invalid county
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('1630615123458')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# The Romanian counties
_COUNTIES = {
    '01': 'Alba',
    '02': 'Arad',
    '03': 'Arges',
    '04': 'Bacau',
    '05': 'Bihor',
    '06': 'Bistrita-Nasaud',
    '07': 'Botosani',
    '08': 'Brasov',
    '09': 'Braila',
    '10': 'Buzau',
    '11': 'Caras-Severin',
    '12': 'Cluj',
    '13': 'Constanta',
    '14': 'Covasna',
    '15': 'Dambovita',
    '16': 'Dolj',
    '17': 'Galati',
    '18': 'Gorj',
    '19': 'Harghita',
    '20': 'Hunedoara',
    '21': 'Ialomita',
    '22': 'Iasi',
    '23': 'Ilfov',
    '24': 'Maramures',
    '25': 'Mehedinti',
    '26': 'Mures',
    '27': 'Neamt',
    '28': 'Olt',
    '29': 'Prahova',
    '30': 'Satu Mare',
    '31': 'Salaj',
    '32': 'Sibiu',
    '33': 'Suceava',
    '34': 'Teleorman',
    '35': 'Timis',
    '36': 'Tulcea',
    '37': 'Vaslui',
    '38': 'Valcea',
    '39': 'Vrancea',
    '40': 'Bucuresti',
    '41': 'Bucuresti - Sector 1',
    '42': 'Bucuresti - Sector 2',
    '43': 'Bucuresti - Sector 3',
    '44': 'Bucuresti - Sector 4',
    '45': 'Bucuresti - Sector 5',
    '46': 'Bucuresti - Sector 6',
    '47': 'Bucuresti - Sector 7 (desfiintat)',
    '48': 'Bucuresti - Sector 8 (desfiintat)',
    '51': 'Calarasi',
    '52': 'Giurgiu',
}


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def calc_check_digit(number):
    """Calculate the check digit for personal codes."""
    # note that this algorithm has not been confirmed by an independent source
    weights = (2, 7, 9, 1, 4, 6, 3, 5, 8, 2, 7, 9)
    check = sum(w * int(n) for w, n in zip(weights, number)) % 11
    return '1' if check == 10 else str(check)


def get_birth_date(number):
    """Split the date parts from the number and return the birth date."""
    number = compact(number)
    centuries = {
        '1': 1900, '2': 1900, '3': 1800, '4': 1800, '5': 2000, '6': 2000,
    }  # we assume 1900 for the others in order to try to construct a date
    year = int(number[1:3]) + centuries.get(number[0], 1900)
    month = int(number[3:5])
    day = int(number[5:7])
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_county(number):
    """Get the county name from the number"""
    try:
        return _COUNTIES[compact(number)[7:9]]
    except KeyError:
        raise InvalidComponent()


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    # first digit should be a known one
    # (7,8=foreign resident, 9=other foreigner but apparently only as NIF)
    if number[0] not in '123456789':
        raise InvalidComponent()
    if len(number) != 13:
        raise InvalidLength()
    # check if birth date is valid
    get_birth_date(number)
    # TODO: check that the birth date is not in the future
    get_county(number)
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import json
import os
import requests

from typing import Dict
from requests import Response

from helper import retry, json_serialize

from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import ErrorDownloadingBinaries
from application.util import LoggerUtil, ErrorResponseUtil

logger = LoggerUtil.get_logger_name(__name__)


class BinaryUpdaterAPIRepository(object):
    def __init__(self, configs):
        # type: (Configurations) -> None

        self.configs = configs
        self.update_type = UpdateType.binaries
        self.local_configs = configs.updaters[self.update_type.name]
        self.response_util = ErrorResponseUtil()
        self.error_response = None

    def check_last_update_binaries(self):
        # type: () -> str

        response_backoffice = self.make_request(self._get_backoffice_url())
        if response_backoffice is None:
            raise ErrorDownloadingBinaries(self.error_response)
        return json.loads(response_backoffice.text)

    def download_src(self, src_version_link, file_src_path):
        # type: (str, str) -> None

        logger.info("Downloading src")
        response_src = self.make_request_zip(src_version_link)
        self._save_package(file_src_path, response_src)

    def download_core(self, core_version_link, file_core_path):
        # type: (str, str) -> None

        logger.info("Downloading core")
        response_core = self.make_request_zip(core_version_link)
        self._save_package(file_core_path, response_core)

    @staticmethod
    def save_file(file_path, file_content):
        # type: (str, str) -> None

        with open(file_path, "w") as object:
            object.write(file_content)

    @staticmethod
    def delete_file(file_path):
        # type: (str) -> None

        has_file = os.path.exists(file_path)
        if has_file:
            os.remove(file_path)

    def notify_update(
        self,           # type: BinaryUpdaterAPIRepository
        version_id,     # type: str
    ):
        # type: (str) -> bool

        response = self.make_request(self._get_backoffice_url() + "/" + version_id, "PUT")
        return response is not None

    def make_request(self, url, method="GET"):
        # type: (str, str) -> Response

        headers = self._get_json_headers()
        response = retry(
            logger=logger,
            method_name="backoffice",
            n=3,
            sleep_seconds=5,
            trying_to_do=lambda: requests.request(method, url=url, headers=headers, timeout=60),
            is_success=lambda resp: resp.status_code == 200,
            success=lambda resp: resp,
            failed=lambda resp: self._get_error_response(resp))

        return response

    def make_request_zip(self, url):
        # type: (str) -> Response

        headers = self._get_tar_gz_headers()
        response = retry(
            logger=logger,
            method_name="download_binaries",
            n=3,
            sleep_seconds=5,
            trying_to_do=lambda: requests.request("GET", url=url, headers=headers, timeout=60),
            is_success=lambda resp: resp.status_code == 200,
            success=lambda resp: resp,
            failed=lambda resp: self._get_error_response(resp))

        if response is None:
            raise ErrorDownloadingBinaries(self.error_response)

        return response

    @staticmethod
    def _save_package(file_path, package):
        # type: (str, Response) -> None

        with open(file_path, "wb") as zip_object:
            for chunk in package.iter_content(chunk_size=128):
                zip_object.write(chunk)

    @staticmethod
    def _download_package(download_url):
        # type: (str) -> Response

        package = requests.get(download_url, allow_redirects=True)
        return package

    def _get_backoffice_url(self):
        # type: () -> str

        end_point_url = self.local_configs.endpoints.backoffice.format(storeId=self.configs.store_id)
        return self.configs.base_url + end_point_url

    def _get_json_headers(self):
        # type: () -> Dict[str, str]

        return {"X-Api-Key": self.configs.api_key}

    @staticmethod
    def _get_tar_gz_headers():
        # type: () -> Dict[str, str]

        return {"Content-type": "application/x-tar"}

    def _get_error_response(self, resp):
        # type: (str) -> None

        logger.error("Error downloading Binaries: {}".format(json_serialize(resp)))
        self.error_response = self.response_util.get_error_response(resp, "$BINARIES_NOT_FOUND")



import iso8601
from typing import (
    Dict,
    List,
)

from devicemanager.model import (
    DeviceStatus,
    DeviceStatusHistory,
)
from devicemanager.repository import DeviceManagerRepository
from externals.persist import ConnectionWrapper


class SqliteDeviceManagerRepository(DeviceManagerRepository):

    def __init__(
        self,
        conn,   # type: ConnectionWrapper
    ):
        self.conn = conn

    def get_disabled_devices(
        self,   # type: SqliteDeviceManagerRepository
    ):
        # type: (...) -> Dict[str, DeviceStatus]

        query = """
            SELECT DeviceName 
            FROM DisabledDevices
        """
        cursor = self.conn.select(sql=query)

        ret = {}
        for row in cursor:
            device = DeviceStatus(
                device_name=row.get_entry("DeviceName").decode("utf-8"),
                enabled=False,
            )
            ret[device.device_name] = device
        return ret

    def get_devices_with_last_history(
        self,           # type: SqliteDeviceManagerRepository
        device_names,   # type: List[str]
    ):
        # type: (...) -> List[DeviceStatus]

        query = """
            SELECT 
                dh.DeviceName, 
                dh.Operator, 
                dh.Timestamp, 
                dd.DeviceName IS NOT NULL AS Enabled
            FROM 
            (
                SELECT 
                    DeviceName, 
                    Operator, 
                    MAX(Timestamp ) AS Timestamp
                FROM DeviceStatusHistory 
                WHERE DeviceName IN ({})
                GROUP BY DeviceName
            ) dh 
            LEFT OUTER JOIN DisabledDevices dd on dh.DeviceName = dd.DeviceName
        """.format("'{}'".format("','".join(device_names)))

        cursor = self.conn.select(sql=query)

        ret = []
        for row in cursor:
            enabled = int(row.get_entry("Enabled")) == 0
            device = DeviceStatus(
                device_name=row.get_entry("DeviceName").decode("utf-8"),
                enabled=enabled,
                history=[
                    DeviceStatusHistory(
                        enabled=enabled,
                        operator=row.get_entry("Operator").decode("utf-8"),
                        timestamp=iso8601.parse_date(row.get_entry("Timestamp")),
                    )
                ]
            )
            ret.append(device)
        return ret

    def create_device_status(
        self,           # type: SqliteDeviceManagerRepository
        device_name,    # type: str
    ):
        # type: (...) -> None

        query = """
            INSERT INTO DisabledDevices (DeviceName) 
            VALUES ('{}')
        """.format(device_name)
        self.conn.query(sql=query)

    def delete_device_status(
        self,           # type: SqliteDeviceManagerRepository
        device_name,    # type: str
    ):
        # type: (...) -> None

        query = """
            DELETE FROM DisabledDevices 
            WHERE DeviceName = '{}'
        """.format(device_name)
        self.conn.query(sql=query)

    def create_device_status_history(
        self,                   # type: SqliteDeviceManagerRepository
        device_name,            # type: str
        device_status_history,  # type: DeviceStatusHistory
    ):
        # type: (...) -> None

        query = """
            INSERT INTO DeviceStatusHistory (
              DeviceName, 
              Enabled,
              Operator,
              Timestamp
            ) VALUES (
             '{}', 
             '{}',
             '{}',
             '{}'
             )
        """.format(
            device_name,
            str(device_status_history.enabled).lower(),
            device_status_history.operator.encode("utf-8"),
            device_status_history.timestamp.isoformat(),
        )

        self.conn.query(sql=query)



# -*- coding: utf-8 -*-
from logging import Logger

from cfgtools import Group
from modelcommon.model import SitefResponse

from dialogmanager import DialogManager
from mbcontextmessagehandler import MbContextMessageBus
from typing import (
    Any,
    Union,
    Optional,
)

from paymentprocessor import PaymentProcessor
from _CapptaProcessor import CapptaProcessor
from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    TestConnectionData,
    FinishPaymentData,
)
from util import Cache


class CapptaCreditPaymentProcessor(PaymentProcessor):
    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        dialog_manager,     # type: DialogManager
        config,             # type: Group
        logger,             # type: Logger
        cache,              # type: Cache
    ):
        # type: (...) -> None

        super(CapptaCreditPaymentProcessor, self).__init__(message_bus, dialog_manager, config, logger, cache)
        self.cappta_processor = CapptaProcessor(message_bus, dialog_manager, config, logger, cache)

    def connection_status(
        self,
        test_connection_data    # type: TestConnectionData
    ):
        # type: (...) -> bool

        return self.cappta_processor.connection_status(test_connection_data=test_connection_data)

    def add_payment(self, start_payment_data):
        return self.cappta_processor.add_payment(
            start_payment_data=start_payment_data,
        )

    def cancel_payment_order(
        self,
        cancel_payment_order_data,  # type: CancelPaymentOrderData
        additional_params="",       # type: Optional[str]
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        return self.cappta_processor.cancel_payment_order(cancel_payment_order_data=cancel_payment_order_data)

    def cancel_payment(self):
        return self.cappta_processor.cancel_payment()

    def finish_payment(
        self,
        finish_payment_data,        # type: FinishPaymentData
    ):
        # type: (...) -> bool

        return self.cappta_processor.finish_payment(
            finish_payment_data=finish_payment_data,
        )

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        self.cappta_processor.confirm_finish_payment(confirm_finish_payment_data=confirm_finish_payment_data)

    def get_name(self):
        return self.cappta_processor.get_name()

    def get_document(
        self,
        pos_id      # type: str
    ):
        # type: (...) -> str

        return self.cappta_processor.get_document(pos_id)

    def stop(self):
        self.cappta_processor.stop()

    def get_last_receipt(
        self,
        eft_processor,    # type: str
    ):
        # type: (...) -> Optional[str]

        return self.cappta_processor.get_last_receipt(eft_processor=eft_processor)



from production.model import TagEventType
from ._OrderChangerProductionBox import OrderChangerProductionBox


class LastTagTimerBox(OrderChangerProductionBox):

    def __init__(self, name, sons, tag, logger=None):
        super(LastTagTimerBox, self).__init__(name, sons, logger)
        self.tag = tag

    def change_order(self, order):
        last_timestamp = self.get_last_timestamp(order)
        if last_timestamp is not None:
            order.display_time = last_timestamp
        return order

    def get_last_timestamp(self, order):
        for item in order.items:
            if item.has_tag(self.tag):
                for history in item.tag_history:
                    if history.tag == self.tag and history.action == TagEventType.added:
                        return history.date.isoformat()

        return None



# ncf.py - functions for handling Dominican Republic invoice numbers
# coding: utf-8
#
# Copyright (C) 2017-2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

# Development of this functionality was funded by iterativo | https://iterativo.do

"""NCF (Números de Comprobante Fiscal, Dominican Republic receipt number).

The NCF is used to number invoices and other documents for the purpose of tax
filing. The e-CF (Comprobante Fiscal Electrónico) is used together with a
digital certificate for the same purpose. The number is either 19, 11 or 13
(e-CF) digits long.

The 19 digit number starts with a letter (A or P) to indicate that the number
was assigned by the taxpayer or the DGII, followed a 2-digit business unit
number, a 3-digit location number, a 3-digit mechanism identifier, a 2-digit
document type and a 8-digit serial number.

The 11 digit number always starts with a B followed a 2-digit document type
and a 7-digit serial number.

The 13 digit e-CF starts with an E followed a 2-digit document type and an
8-digit serial number.

More information:

 * https://www.dgii.gov.do/
 * https://dgii.gov.do/workshopProveedoresTI-eCE/Documents/Norma05-19.pdf
 * https://dgii.gov.do/cicloContribuyente/facturacion/comprobantesFiscales/Paginas/tiposComprobantes.aspx

>>> validate('E310000000005')  # format since 2019-04-08
'E310000000005'
>>> validate('B0100000005')  # format since 2018-05-01
'B0100000005'
>>> validate('A020010210100000005')  # format before 2018-05-01
'A020010210100000005'
>>> validate('Z0100000005')
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip().upper()


# The following document types are known:
_ncf_document_types = (
    '01',  # invoices for fiscal declaration (or tax reporting)
    '02',  # invoices for final consumer
    '03',  # debit note
    '04',  # credit note (refunds)
    '11',  # informal supplier invoices (purchases)
    '12',  # single income invoices
    '13',  # minor expenses invoices (purchases)
    '14',  # invoices for special customers (tourists, free zones)
    '15',  # invoices for the government
    '16',  # invoices for export
    '17',  # invoices for payments abroad
)

_ecf_document_types = (
    '31',  # invoices for fiscal declaration (or tax reporting)
    '32',  # invoices for final consumer
    '33',  # debit note
    '34',  # credit note (refunds)
    '41',  # supplier invoices (purchases)
    '43',  # minor expenses invoices (purchases)
    '44',  # invoices for special customers (tourists, free zones)
    '45',  # invoices for the government
    '46',  # invoices for exports
    '47',  # invoices for foreign payments
)


def validate(number):
    """Check if the number provided is a valid NCF."""
    number = compact(number)
    if len(number) == 13:
        if number[0] != 'E' or not isdigits(number[1:]):
            raise InvalidFormat()
        if number[1:3] not in _ecf_document_types:
            raise InvalidComponent()
    elif len(number) == 11:
        if number[0] != 'B' or not isdigits(number[1:]):
            raise InvalidFormat()
        if number[1:3] not in _ncf_document_types:
            raise InvalidComponent()
    elif len(number) == 19:
        if number[0] not in 'AP' or not isdigits(number[1:]):
            raise InvalidFormat()
        if number[9:11] not in _ncf_document_types:
            raise InvalidComponent()
    else:
        raise InvalidLength()
    return number


def is_valid(number):
    """Check if the number provided is a valid NCF."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def _convert_result(result):  # pragma: no cover
    """Translate SOAP result entries into dictionaries."""
    translation = {
        'NOMBRE': 'name',
        'COMPROBANTE': 'proof',
        'ES_VALIDO': 'is_valid',
        'MENSAJE_VALIDACION': 'validation_message',
        'RNC': 'rnc',
        'NCF': 'ncf',
        u'RNC / Cédula': 'rnc',
        u'RNC/Cédula': 'rnc',
        u'Nombre / Razón Social': 'name',
        u'Nombre/Razón Social': 'name',
        'Estado': 'status',
        'Tipo de comprobante': 'type',
        u'Válido hasta': 'valid_until',
        u'Código de Seguridad': 'security_code',
        'Rnc Emisor': 'issuing_rnc',
        'Rnc Comprador': 'buyer_rnc',
        'Monto Total': 'total',
        'Total de ITBIS': 'total_itbis',
        'Fecha Emisi&oacuten': 'issuing_date',
        u'Fecha Emisión': 'issuing_date',
        u'Fecha de Firma': 'signature_date',
        'e-NCF': 'ncf',
    }
    return dict(
        (translation.get(key, key), value)
        for key, value in result.items())


def check_dgii(rnc, ncf, buyer_rnc=None, security_code=None, timeout=30):  # pragma: no cover
    """Validate the RNC, NCF combination on using the DGII online web service.

    This uses the validation service run by the the Dirección General de
    Impuestos Internos, the Dominican Republic tax department to check
    whether the combination of RNC and NCF is valid. The timeout is in
    seconds.

    Returns a dict with the following structure for a NCF::

        {
            'name': 'The registered name',
            'status': 'VIGENTE',
            'type': 'FACTURAS DE CREDITO FISCAL',
            'rnc': '123456789',
            'ncf': 'A020010210100000005',
            'validation_message': 'El NCF digitado es válido.',
        }

    For an ECNF::

        {
            'status': 'Aceptado',
            'issuing_rnc': '1234567890123',
            'buyer_rnc': '123456789',
            'ncf': 'E300000000000',
            'security_code': '1+2kP3',
            'issuing_date': '2020-03-25',
            'signature_date': '2020-03-22',
            'total': '2203.50',
            'total_itbis': '305.10',
            'validation_message': 'Aceptado',
        }

    Will return None if the number is invalid or unknown."""
    import lxml.html
    import requests
    from stdnum.do.rnc import compact as rnc_compact  # noqa: I003
    rnc = rnc_compact(rnc)
    ncf = compact(ncf)
    if buyer_rnc:
        buyer_rnc = rnc_compact(buyer_rnc)
    url = 'https://dgii.gov.do/app/WebApps/ConsultasWeb2/ConsultasWeb/consultas/ncf.aspx'
    session = requests.Session()
    session.headers.update({
        'User-Agent': 'Mozilla/5.0 (python-stdnum)',
    })
    # Get the page to pick up needed form parameters
    document = lxml.html.fromstring(
        session.get(url, timeout=timeout).text)
    validation = document.find('.//input[@name="__EVENTVALIDATION"]').get('value')
    viewstate = document.find('.//input[@name="__VIEWSTATE"]').get('value')
    data = {
        '__EVENTVALIDATION': validation,
        '__VIEWSTATE': viewstate,
        'ctl00$cphMain$btnConsultar': 'Buscar',
        'ctl00$cphMain$txtNCF': ncf,
        'ctl00$cphMain$txtRNC': rnc,
    }
    if ncf[0] == 'E':
        data['ctl00$cphMain$txtRncComprador'] = buyer_rnc
        data['ctl00$cphMain$txtCodigoSeg'] = security_code
    # Do the actual request
    document = lxml.html.fromstring(
        session.post(url, data=data, timeout=timeout).text)
    result_path = './/div[@id="cphMain_PResultadoFE"]' if ncf[0] == 'E' else './/div[@id="cphMain_pResultado"]'
    result = document.find(result_path)
    if result is not None:
        lbl_path = './/*[@id="cphMain_lblEstadoFe"]' if ncf[0] == 'E' else './/*[@id="cphMain_lblInformacion"]'
        data = {
            'validation_message': document.findtext(lbl_path).strip(),
        }
        data.update(zip(
            [x.text.strip() for x in result.findall('.//th') if x.text],
            [x.text.strip() for x in result.findall('.//td/span') if x.text]))
        return _convert_result(data)



from sysactions import get_storewide_config

must_print_delivery_coupon_config = None


def must_print_delivery_coupon():
    # type: (...) -> bool

    global must_print_delivery_coupon_config

    if must_print_delivery_coupon_config is None:
        config = get_storewide_config("Store.DeliveryConfig.PrintDeliveryCoupon", defval="true")
        must_print_delivery_coupon_config = config.lower() == "true"

    return must_print_delivery_coupon_config



# -*- coding: utf-8 -*-
"""
    flask.debughelpers
    ~~~~~~~~~~~~~~~~~~

    Various helpers to make the development experience better.

    :copyright: 2010 Pallets
    :license: BSD-3-Clause
"""
import os
from warnings import warn

from ._compat import implements_to_string
from ._compat import text_type
from .app import Flask
from .blueprints import Blueprint
from .globals import _request_ctx_stack


class UnexpectedUnicodeError(AssertionError, UnicodeError):
    """Raised in places where we want some better error reporting for
    unexpected unicode or binary data.
    """


@implements_to_string
class DebugFilesKeyError(KeyError, AssertionError):
    """Raised from request.files during debugging.  The idea is that it can
    provide a better error message than just a generic KeyError/BadRequest.
    """

    def __init__(self, request, key):
        form_matches = request.form.getlist(key)
        buf = [
            'You tried to access the file "%s" in the request.files '
            "dictionary but it does not exist.  The mimetype for the request "
            'is "%s" instead of "multipart/form-data" which means that no '
            "file contents were transmitted.  To fix this error you should "
            'provide enctype="multipart/form-data" in your form.'
            % (key, request.mimetype)
        ]
        if form_matches:
            buf.append(
                "\n\nThe browser instead transmitted some file names. "
                "This was submitted: %s" % ", ".join('"%s"' % x for x in form_matches)
            )
        self.msg = "".join(buf)

    def __str__(self):
        return self.msg


class FormDataRoutingRedirect(AssertionError):
    """This exception is raised by Flask in debug mode if it detects a
    redirect caused by the routing system when the request method is not
    GET, HEAD or OPTIONS.  Reasoning: form data will be dropped.
    """

    def __init__(self, request):
        exc = request.routing_exception
        buf = [
            "A request was sent to this URL (%s) but a redirect was "
            'issued automatically by the routing system to "%s".'
            % (request.url, exc.new_url)
        ]

        # In case just a slash was appended we can be extra helpful
        if request.base_url + "/" == exc.new_url.split("?")[0]:
            buf.append(
                "  The URL was defined with a trailing slash so "
                "Flask will automatically redirect to the URL "
                "with the trailing slash if it was accessed "
                "without one."
            )

        buf.append(
            "  Make sure to directly send your %s-request to this URL "
            "since we can't make browsers or HTTP clients redirect "
            "with form data reliably or without user interaction." % request.method
        )
        buf.append("\n\nNote: this exception is only raised in debug mode")
        AssertionError.__init__(self, "".join(buf).encode("utf-8"))


def attach_enctype_error_multidict(request):
    """Since Flask 0.8 we're monkeypatching the files object in case a
    request is detected that does not use multipart form data but the files
    object is accessed.
    """
    oldcls = request.files.__class__

    class newcls(oldcls):
        def __getitem__(self, key):
            try:
                return oldcls.__getitem__(self, key)
            except KeyError:
                if key not in request.form:
                    raise
                raise DebugFilesKeyError(request, key)

    newcls.__name__ = oldcls.__name__
    newcls.__module__ = oldcls.__module__
    request.files.__class__ = newcls


def _dump_loader_info(loader):
    yield "class: %s.%s" % (type(loader).__module__, type(loader).__name__)
    for key, value in sorted(loader.__dict__.items()):
        if key.startswith("_"):
            continue
        if isinstance(value, (tuple, list)):
            if not all(isinstance(x, (str, text_type)) for x in value):
                continue
            yield "%s:" % key
            for item in value:
                yield "  - %s" % item
            continue
        elif not isinstance(value, (str, text_type, int, float, bool)):
            continue
        yield "%s: %r" % (key, value)


def explain_template_loading_attempts(app, template, attempts):
    """This should help developers understand what failed"""
    info = ['Locating template "%s":' % template]
    total_found = 0
    blueprint = None
    reqctx = _request_ctx_stack.top
    if reqctx is not None and reqctx.request.blueprint is not None:
        blueprint = reqctx.request.blueprint

    for idx, (loader, srcobj, triple) in enumerate(attempts):
        if isinstance(srcobj, Flask):
            src_info = 'application "%s"' % srcobj.import_name
        elif isinstance(srcobj, Blueprint):
            src_info = 'blueprint "%s" (%s)' % (srcobj.name, srcobj.import_name)
        else:
            src_info = repr(srcobj)

        info.append("% 5d: trying loader of %s" % (idx + 1, src_info))

        for line in _dump_loader_info(loader):
            info.append("       %s" % line)

        if triple is None:
            detail = "no match"
        else:
            detail = "found (%r)" % (triple[1] or "<string>")
            total_found += 1
        info.append("       -> %s" % detail)

    seems_fishy = False
    if total_found == 0:
        info.append("Error: the template could not be found.")
        seems_fishy = True
    elif total_found > 1:
        info.append("Warning: multiple loaders returned a match for the template.")
        seems_fishy = True

    if blueprint is not None and seems_fishy:
        info.append(
            "  The template was looked up from an endpoint that "
            'belongs to the blueprint "%s".' % blueprint
        )
        info.append("  Maybe you did not place a template in the right folder?")
        info.append("  See http://flask.pocoo.org/docs/blueprints/#templates")

    app.logger.info("\n".join(info))


def explain_ignored_app_run():
    if os.environ.get("WERKZEUG_RUN_MAIN") != "true":
        warn(
            Warning(
                "Silently ignoring app.run() because the "
                "application is run from the flask command line "
                "executable.  Consider putting app.run() behind an "
                'if __name__ == "__main__" guard to silence this '
                "warning."
            ),
            stacklevel=3,
        )



u"""
Fixer for memoryview(s) -> buffer(s).
Explicit because some memoryview methods are invalid on buffer objects.
"""

from lib2to3 import fixer_base
from lib2to3.fixer_util import Name


class FixMemoryview(fixer_base.BaseFix):

    explicit = True # User must specify that they want this.

    PATTERN = u"""
              power< name='memoryview' trailer< '(' [any] ')' >
              rest=any* >
              """

    def transform(self, node, results):
        name = results[u"name"]
        name.replace(Name(u"buffer", prefix=name.prefix))



import sysactions

from pos_model import Order
from xml.etree import cElementTree as eTree
from helper import remove_accents


class InfAdicBuilderCommon(object):
    def __init__(self, order, customer_text_field, tax_builder):
        self.customer_text_field = customer_text_field
        self.tax_builder = tax_builder
        self.order = order

    def build_common(self):
        # type: () -> unicode

        return u"<infAdic><infCpl>{0}</infCpl></infAdic>".format(self._get_inf_cpl())

    def _get_inf_cpl(self):
        # type: () -> unicode

        xml = self._get_order_identifier(self.order.order_picture_xml)
        xml += self._get_tax_data(self.order.order_picture_xml)
        xml += remove_accents(self.customer_text_field.replace('\n', ''))

        return xml.strip()

    def _get_tax_data(self, order):
        # type: (Order) -> unicode

        total_taxes = self.tax_builder.ibpt_tax_calculator(order)
        total_tax_amount = float(total_taxes['nacionalfederal']['value']) + float(total_taxes['estadual']['value'])
        xml = "%-44s R$%7.2f - " % ("Valor aproximado dos tributos deste cupom", total_tax_amount)
        xml += "Fed. R$%7.2f (%.2f%%) Est. R$%7.2f (%.2f%%)" % (total_taxes['nacionalfederal']['value'],
                                                                total_taxes['nacionalfederal']['percent'],
                                                                total_taxes['estadual']['value'],
                                                                total_taxes['estadual']['percent'])
        xml += " (Conforme Lei Fed. 12.741/2012)"

        return remove_accents(xml)

    def _get_order_identifier(
        self,   # type: InfAdicBuilderCommon
        order,  # type: eTree.Element
    ):
        # type: (...) -> unicode

        order_id = order.get("orderId")
        user_id = order.get("sessionId").split("user=")[1].split(",")[0]
        user_info = sysactions.get_user_information(user_number=user_id)

        operator_name = user_id
        if user_info:
            user = eTree.XML(user_info).find(".//user")
            operator_name = user.get("LongName")[:15] if user is not None else user_id

        table_property = order.find("CustomOrderProperties/OrderProperty[@key='TABLE_ID']")
        table_id = table_property.get("value") if table_property is not None else None
        if table_id:
            order_label = "Pedido/Mesa: {}/{}".format(order_id[-4:], table_id)
        else:
            order_label = "Pedido: {}".format(order_id[-4:])

        xml = "Operador: {}  {} ".format(operator_name, order_label)
        xml += self._get_delivery_partner_and_short_reference(order)

        return remove_accents(xml)

    @staticmethod
    def _get_delivery_partner_and_short_reference(
        order,  # type: eTree.Element
    ):
        # type: (...) -> str

        delivery_partner = order.find("CustomOrderProperties/OrderProperty[@key='PARTNER']")
        if delivery_partner is not None:
            short_reference = order.find("CustomOrderProperties/OrderProperty[@key='SHORT_REFERENCE']")
            if short_reference is not None:
                return "{}: {} ".format(delivery_partner.get("value").upper(), short_reference.get("value").upper())

        return " "



import decimal
import hmac
import numbers
import sys

PY2 = sys.version_info[0] == 2

if PY2:
    from itertools import izip

    text_type = unicode  # noqa: 821
else:
    izip = zip
    text_type = str

number_types = (numbers.Real, decimal.Decimal)


def _constant_time_compare(val1, val2):
    """Return ``True`` if the two strings are equal, ``False``
    otherwise.

    The time taken is independent of the number of characters that
    match. Do not use this function for anything else than comparision
    with known length targets.

    This is should be implemented in C in order to get it completely
    right.

    This is an alias of :func:`hmac.compare_digest` on Python>=2.7,3.3.
    """
    len_eq = len(val1) == len(val2)
    if len_eq:
        result = 0
        left = val1
    else:
        result = 1
        left = val2
    for x, y in izip(bytearray(left), bytearray(val2)):
        result |= x ^ y
    return result == 0


# Starting with 2.7/3.3 the standard library has a c-implementation for
# constant time string compares.
constant_time_compare = getattr(hmac, "compare_digest", _constant_time_compare)



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Optional  # noqa


class InfoMessage(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def show(self, message, msg_type=None):
        # type: (str, Optional[InfoMessage.Type]) -> None
        raise NotImplementedError()

    class Type(Enum):
        info = 1
        warn = 2
        error = 3



"""
General functions for HTML manipulation, backported from Py3.

Note that this uses Python 2.7 code with the corresponding Python 3
module names and locations.
"""

from __future__ import unicode_literals


_escape_map = {ord('&'): '&amp;', ord('<'): '&lt;', ord('>'): '&gt;'}
_escape_map_full = {ord('&'): '&amp;', ord('<'): '&lt;', ord('>'): '&gt;',
                    ord('"'): '&quot;', ord('\''): '&#x27;'}

# NB: this is a candidate for a bytes/string polymorphic interface

def escape(s, quote=True):
    """
    Replace special characters "&", "<" and ">" to HTML-safe sequences.
    If the optional flag quote is true (the default), the quotation mark
    characters, both double quote (") and single quote (') characters are also
    translated.
    """
    assert not isinstance(s, bytes), 'Pass a unicode string'
    if quote:
        return s.translate(_escape_map_full)
    return s.translate(_escape_map)

